# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

from hamcrest import assert_that, calling, equal_to, raises
import mock
from requests import exceptions
import tempfile

from storops.connection import client
from storops.connection import exceptions as storops_ex
from storops.exception import StoropsConnectTimeoutError


class FakeOpener(dict):
    def __eq__(self, other):
        if 'filename' in self and 'filename' in other:
            return other['filename'].name == self['filename'].name
        return False


class MockResponse(object):
    def __init__(self, text, status_code):
        self.text = text
        self.status_code = status_code


def _request_side_effect(method, url, **kwargs):
    response_map = {'right_url': MockResponse('OK', 200),
                    'right_url_json': MockResponse('{"id": "id_123"}', 200),
                    'bad_url': MockResponse('Failed', 404),
                    'bad_url_raise': MockResponse('Failed', 401),
                    'service_unavailable':
                        MockResponse('Service unavailable', 503)}
    return response_map[url]


class ClientModuleTest(unittest.TestCase):
    def test_on_error_callback_true(self):
        result = client._on_error_callback(exceptions.RequestException())
        assert_that(True, equal_to(result))

    def test_on_error_callback_false(self):
        result = client._on_error_callback(ValueError())
        assert_that(False, equal_to(result))

    def test_wait_callback(self):
        assert_that(8, equal_to(client._wait_callback(4)))


class HTTPClientTest(unittest.TestCase):

    def setUp(self):
        self.client = client.HTTPClient('https://10.10.10.10', {})

    def test_set_request_options_insecure_true(self):
        options = client.HTTPClient._set_request_options(insecure=False,
                                                         auth=None,
                                                         timeout=None,
                                                         ca_cert_path=None)
        assert_that(options, equal_to({'verify': True,
                                       'auth': None}))

    def test_set_request_options_insecure_true_path(self):
        options = client.HTTPClient._set_request_options(
            insecure=False, auth=None, timeout=None, ca_cert_path='/tmp.crt')
        assert_that(options, equal_to({'verify': '/tmp.crt',
                                       'auth': None}))

    def test_set_request_options_insecure_false(self):
        options = client.HTTPClient._set_request_options(
            insecure=True, auth=None, timeout=None, ca_cert_path='/tmp.crt')
        assert_that(options, equal_to({'verify': False,
                                       'auth': None}))

    def test_request_content_json(self):
        self.client.session.request = mock.MagicMock(
            side_effect=_request_side_effect)
        self.client.headers['Content-Type'] = 'application/json'

        resp, body = self.client.request('right_url_json', 'GET',
                                         body={"k_abc": "v_abc"})

        self.client.session.request.assert_called_with(
            'GET', 'right_url_json', auth=None, files=None, verify=True,
            headers={'Content-Type': 'application/json'},
            data='{"k_abc": "v_abc"}')
        assert_that(resp.status_code, equal_to(200))
        assert_that(body, equal_to({'id': 'id_123'}))

    def test_request_with_file(self):
        self.client.session.request = mock.MagicMock(
            side_effect=_request_side_effect)
        self.client.headers['Content-Type'] = 'application/json'
        _, name = tempfile.mkstemp(suffix='.lic')
        files = dict()
        files['filename'] = name
        resp, body = self.client.request('right_url_json', 'GET',
                                         files=files, body={"k_abc": "v_abc"})

        fake_files = FakeOpener()
        fake_files['filename'] = open(name, 'rb')
        self.client.session.request.assert_called_with(
            'GET', 'right_url_json', auth=None, files=fake_files, verify=True,
            headers={}, data='{"k_abc": "v_abc"}')

        assert_that(resp.status_code, equal_to(200))
        assert_that(body, equal_to({'id': 'id_123'}))

    def test_request_content_plain(self):
        self.client.session.request = mock.MagicMock(
            side_effect=_request_side_effect)

        resp, body = self.client.request('right_url', 'GET',
                                         body='{"k_abc": "v_abc"}')

        self.client.session.request.assert_called_with(
            'GET', 'right_url', auth=None, files=None,
            verify=True, headers={},
            data='{"k_abc": "v_abc"}')
        assert_that(resp.status_code, equal_to(200))
        assert_that(body, equal_to('OK'))

    def test_request_content_404(self):
        self.client.session.request = mock.MagicMock(
            side_effect=_request_side_effect)

        resp, body = self.client.request('bad_url', 'GET',
                                         body='{"k_abc": "v_abc"}')

        self.client.session.request.assert_called_with(
            'GET', 'bad_url', auth=None, files=None,
            verify=True, headers={}, data='{"k_abc": "v_abc"}')
        assert_that(resp.status_code, equal_to(404))
        assert_that(body, equal_to('Failed'))

    @mock.patch('storops.connection.exceptions.from_response')
    def test_request_content_raise(self, mocked_from_response):
        self.client.session.request = mock.MagicMock(
            side_effect=_request_side_effect)
        mocked_from_response.return_value = storops_ex.HttpError()

        def _tmp_func():
            self.client.request('bad_url_raise', 'GET',
                                body='{"k_abc": "v_abc"}')
        assert_that(calling(_tmp_func),
                    raises(storops_ex.HttpError))

        self.client.session.request.assert_called_with(
            'GET', 'bad_url_raise', auth=None, files=None,
            verify=True, headers={}, data='{"k_abc": "v_abc"}')

    @mock.patch('storops.connection.exceptions.from_response')
    def test_request_when_service_unavailable(self, mocked_from_response):
        self.client.session.request = mock.MagicMock(
            side_effect=_request_side_effect)
        mocked_from_response.return_value = storops_ex.HttpServerError()

        def _tmp_func():
            self.client.request('service_unavailable', 'GET',
                                body='{"k_abc": "v_abc"}')
        assert_that(calling(_tmp_func),
                    raises(storops_ex.HttpServerError))

        self.client.session.request.assert_called_with(
            'GET', 'service_unavailable', auth=None, files=None,
            verify=True, headers={}, data='{"k_abc": "v_abc"}')

    def test_request_when_connect_timeout(self):
        self.client.session.request = mock.MagicMock(
            side_effect=exceptions.ConnectTimeout)

        def _tmp_func():
            self.client.request('fake_url_connect_timeout', 'GET',
                                body='{"k_abc": "v_abc"}')
        assert_that(calling(_tmp_func),
                    raises(exceptions.ConnectTimeout))

        self.client.session.request.assert_called_with(
            'GET', 'fake_url_connect_timeout', auth=None, files=None,
            verify=True, headers={}, data='{"k_abc": "v_abc"}')

    @mock.patch(
        'storops.connection.client.HTTPClient._cs_request_with_retries')
    def test_cs_request(self, mocked_cs_request_with_retries):
        self.client.base_url = 'https://10.10.10.10'
        self.client._cs_request('/api/types/instance', 'GET')

        mocked_cs_request_with_retries.assert_called_with(
            'https://10.10.10.10/api/types/instance',
            'GET')

    @mock.patch(
        'storops.connection.client._wait_callback')
    def test_cs_request_connect_timeout(self, mocked_wait_callback):
        mocked_wait_callback.return_value = 0

        self.client.session.request = mock.MagicMock(
            side_effect=exceptions.ConnectTimeout)

        self.client.base_url = 'https://10.10.10.10'
        assert_that(calling(self.client._cs_request).with_args(
            '/api/types/instance', 'GET'), raises(StoropsConnectTimeoutError))

        calls = [
            mock.call('GET', 'https://10.10.10.10/api/types/instance',
                      auth=None, files=None, verify=True, headers={}),
            mock.call('GET', 'https://10.10.10.10/api/types/instance',
                      auth=None, files=None, verify=True, headers={}),
        ]
        self.client.session.request.assert_has_calls(calls)

    def test_get_limit(self):
        self.client.retries = 99
        assert_that(99, equal_to(self.client._get_limit()))

    @mock.patch(
        'storops.connection.client.HTTPClient._cs_request')
    def test_get(self, mocked_cs_request):
        self.client.get('/api/types/instance')

        mocked_cs_request.assert_called_with(
            '/api/types/instance',
            'GET')

    @mock.patch(
        'storops.connection.client.HTTPClient._cs_request')
    def test_post(self, mocked_cs_request):
        self.client.post('/api/types/instance')

        mocked_cs_request.assert_called_with(
            '/api/types/instance',
            'POST')

    @mock.patch(
        'storops.connection.client.HTTPClient._cs_request')
    def test_put(self, mocked_cs_request):
        self.client.put('/api/types/instance')

        mocked_cs_request.assert_called_with(
            '/api/types/instance',
            'PUT')

    @mock.patch(
        'storops.connection.client.HTTPClient._cs_request')
    def test_delete(self, mocked_cs_request):
        self.client.delete('/api/types/instance')

        mocked_cs_request.assert_called_with(
            '/api/types/instance',
            'DELETE')



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

import six

import storops.unity.resource.cifs_server
import storops.unity.resource.filesystem
import storops.unity.resource.snap
from storops.exception import UnityCreateSnapError
from storops.lib.common import instance_cache
from storops.unity.enums import CIFSTypeEnum, ACEAccessTypeEnum, \
    ACEAccessLevelEnum, FilesystemSnapAccessTypeEnum
from storops.unity.resource import UnityResource, UnityResourceList, \
    UnityAttributeResource

__author__ = 'Jay Xu'

log = logging.getLogger(__name__)


class UnityCifsShare(UnityResource):
    @classmethod
    def create(cls, cli, name, fs, path=None, cifs_server=None):
        fs_clz = storops.unity.resource.filesystem.UnityFileSystem
        fs = fs_clz.get(cli, fs).verify()
        sr = fs.storage_resource

        if path is None:
            path = '/'

        if cifs_server is None:
            cifs_server = fs.first_available_cifs_server
        else:
            server_clz = storops.unity.resource.cifs_server.UnityCifsServer
            cifs_server = server_clz.get(cli, cifs_server)

        param = cli.make_body(name=name,
                              path=path,
                              cifsServer=cifs_server)
        resp = sr.modify_fs(cifsShareCreate=[param])
        resp.raise_if_err()
        return UnityCifsShareList(cli=cli, name=name).first_item

    @classmethod
    def create_from_snap(cls, cli, snap, name, path=None, is_read_only=None):
        snap_clz = storops.unity.resource.snap.UnitySnap
        snap = snap_clz.get(cli, snap)

        if path is None:
            path = '/'

        resp = cli.post(cls().resource_class,
                        snap=snap,
                        path=path,
                        name=name,
                        isReadOnly=is_read_only)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    @property
    @instance_cache
    def storage_resource(self):
        fs = self.filesystem
        if fs is not None:
            ret = fs.storage_resource
        else:
            ret = None
        return ret

    def delete(self, async_mode=False):
        if self.type == CIFSTypeEnum.CIFS_SNAPSHOT:
            resp = super(UnityCifsShare, self).delete(async_mode=async_mode)
        else:
            fs = self.filesystem.verify()
            sr = fs.storage_resource
            param = self._cli.make_body(cifsShare=self)
            resp = sr.modify_fs(async_mode=async_mode, cifsShareDelete=[param])
        resp.raise_if_err()
        return resp

    def get_ace_list(self):
        return UnityCifsShareAceList(cli=self._cli, cifs_share=self)

    def enable_ace(self):
        return self.modify(is_ace_enabled=True)

    def disable_ace(self):
        return self.modify(is_ace_enabled=False)

    def _get_domain_user_name(self, domain=None, user=None):
        if domain is None:
            domain = self.cifs_server.domain
        if user is None:
            raise ValueError('username not specified.')
        return r'{}\{}'.format(domain, user)

    def clear_access(self, white_list=None):
        """ clear all ace entries of the share

        :param white_list: list of username whose access entry won't be cleared
        :return: sid list of ace entries removed successfully
        """
        access_entries = self.get_ace_list()
        sid_list = access_entries.sid_list

        if white_list:
            sid_white_list = [UnityAclUser.get_sid(self._cli,
                                                   user,
                                                   self.cifs_server.domain)
                              for user in white_list]
            sid_list = list(set(sid_list) - set(sid_white_list))

        resp = self.delete_ace(sid=sid_list)
        resp.raise_if_err()
        return sid_list

    def add_ace(self, domain=None, user=None, access_level=None):
        if domain is None:
            domain = self.cifs_server.domain
        if user is None:
            raise ValueError('cifs username is not specified.')

        if access_level is None:
            access_level = ACEAccessLevelEnum.FULL
        sid = UnityAclUser.get_sid(self._cli, user=user, domain=domain)
        ace = self._cli.make_body(
            sid=sid,
            accessType=ACEAccessTypeEnum.GRANT,
            accessLevel=access_level
        )

        resp = self.action("setACEs", cifsShareACEs=[ace])
        resp.raise_if_err()
        return resp

    def delete_ace(self, domain=None, user=None, sid=None):
        """ delete ACE for the share

        delete ACE for the share.  User could either supply the domain and
        username or the sid of the user.

        :param domain: domain of the user
        :param user: username
        :param sid: sid of the user or sid list of the user
        :return: REST API response
        """
        if sid is None:
            if domain is None:
                domain = self.cifs_server.domain

            sid = UnityAclUser.get_sid(self._cli, user=user, domain=domain)
        if isinstance(sid, six.string_types):
            sid = [sid]
        ace_list = [self._make_remove_ace_entry(s) for s in sid]

        resp = self.action("setACEs", cifsShareACEs=ace_list)
        resp.raise_if_err()
        return resp

    def _make_remove_ace_entry(self, sid):
        return self._cli.make_body(
            sid=sid,
            accessType=ACEAccessTypeEnum.NONE,
            accessLevel=ACEAccessLevelEnum.FULL)

    def modify(self, is_read_only=None, is_ace_enabled=None, add_ace=None,
               delete_ace=None):
        sr = self.storage_resource
        if sr is None:
            raise ValueError('storage resource for share {} not found.'
                             .format(self.name))

        share_param = self._cli.make_body(
            allow_empty=True,
            isReadOnly=is_read_only,
            isACEEnabled=is_ace_enabled,
            addACE=add_ace,
            deleteAce=delete_ace)
        modify_param = self._cli.make_body(
            allow_empty=True,
            cifsShare=self,
            cifsShareParameters=share_param)
        param = self._cli.make_body(
            allow_empty=True,
            cifsShareModify=[modify_param])

        resp = sr.modify_fs(**param)
        resp.raise_if_err()
        return resp

    @property
    @instance_cache
    def cifs_server(self):
        ret = None
        fs = self.filesystem
        if fs:
            nas_server = fs.nas_server
            if nas_server:
                cifs_servers = nas_server.cifs_server
                if cifs_servers:
                    ret = cifs_servers[0]
        return ret

    def create_snap(self, name=None, fs_access_type=None):
        if fs_access_type is None:
            fs_access_type = FilesystemSnapAccessTypeEnum.PROTOCOL

        if self.type == CIFSTypeEnum.CIFS_SHARE:
            ret = self.filesystem.create_snap(
                name=name, fs_access_type=fs_access_type)
        elif self.type == CIFSTypeEnum.CIFS_SNAPSHOT:
            ret = self.snap.copy(copy_name=name)
        else:
            raise UnityCreateSnapError('do not know how to create snap for '
                                       'cifs share {}, type {}.'
                                       .format(self.name, self.type))
        return ret


class UnityCifsShareList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityCifsShare


class UnityCifsShareAce(UnityAttributeResource):
    pass


class UnityCifsShareAceList(UnityResourceList):
    def __init__(self, cli=None, cifs_share=None, **the_filter):
        super(UnityCifsShareAceList, self).__init__(
            cli=cli, **the_filter)
        self.cifs_share = cifs_share

    def _get_raw_resource(self):
        resp = self.cifs_share.action('getACEs')
        resp.raise_if_err()
        return resp.first_content['cifsShareACEs']

    @property
    def sid_list(self):
        return [ace.sid for ace in self]

    @classmethod
    def get_resource_class(cls):
        return UnityCifsShareAce


class UnityAclUser(UnityResource):
    @classmethod
    def get_sid(cls, cli, user, domain=None):
        resp = cli.type_action(cls().resource_class,
                               'lookupSIDByDomainUser',
                               domainName=domain,
                               userName=user)
        resp.raise_if_err()
        return resp.first_content.get('sid')


class UnityAclUserList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityAclUser



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import re

import six

from storops.lib.ex_decorator_factory import MappedErrorCodeDecoratorFactory, \
    ExceptionListDecoratorFactory

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class StoropsException(Exception):
    """Base Storops Exception

    To correctly use this class, inherit from it and define
    a 'message' property. That message will be formatted
    with the keyword arguments provided to the constructor.

    """
    unknown_message = "An unknown exception occurred."

    # class level error message.
    message = None

    # class level error message template.
    message_template = None

    # default error code used in message template.
    default_code = 500

    def __init__(self, message=None, **kwargs):
        if message is None:
            if self.message_template is not None:
                message = self.message_template
            elif self.__class__.message is not None:
                message = self.__class__.message
            else:
                message = self.unknown_message

        self._message = message
        self._kwargs = kwargs

        super(StoropsException, self).__init__()

    @property
    def message(self):
        ret = self._message
        if isinstance(self._message, six.string_types):
            try:
                if 'code' not in self._kwargs:
                    self._kwargs['code'] = self.default_code

                ret = self._message.format(**self._kwargs)

            except KeyError:
                # kwargs doesn't match a variable in the message
                # log the issue and the kwargs
                log.error(
                    'missing param in format string: "{}"'.format(
                        self._message))
            except IndexError:
                # format error, use original message
                pass
        elif isinstance(self._message, Exception):
            ret = six.text_type(self._message)
        return ret

    @message.setter
    def message(self, value):
        self._message = value

    def __str__(self):
        return self.message


def to_hex(number):
    if number is not None:
        h = hex(number)
        if h.endswith('L'):
            h = h[:-1]
    else:
        h = None
    return h


class VNXException(StoropsException):
    @classmethod
    def get_error_message(cls):
        ret = None
        if hasattr(cls, 'error_code'):
            ret = to_hex(cls.error_code)
        elif hasattr(cls, 'error_message'):
            ret = cls.error_message
        elif hasattr(cls, 'error_regex'):
            flags = re.IGNORECASE | re.MULTILINE | re.DOTALL
            ret = re.compile(cls.error_regex, flags=flags)

        if ret is None:
            raise AttributeError('"error_code" or "error_message" must be '
                                 'specified for VNX CLI error.')
        return ret


class UnityException(StoropsException):
    def __init__(self, error=None):
        super(UnityException, self).__init__()
        self.error = error

    def __str__(self):
        if hasattr(self.error, 'get_messages'):
            ret = '.  '.join(self.error.get_messages())
        elif isinstance(self.error, six.string_types):
            ret = self.error
        elif hasattr(self, 'message'):
            ret = self.message
        else:
            ret = self.error
        return ret

    @property
    def error_code(self):
        if self.error is not None:
            ret = self.error.error_code
        else:
            ret = None
        return ret


class VNXObjectNotFoundError(object):
    pass


class VNXBackendError(VNXException):
    message_template = "vnx backend error.  {err}"


_rest_exception_factory = MappedErrorCodeDecoratorFactory(
    default_exception=UnityException)
get_rest_exception = _rest_exception_factory.get_exception
rest_exception = _rest_exception_factory.clz_decorator()
rest_exception.__doc__ = """ class decorator for Unity REST exceptions

Each Unity REST exception has a designated error code.
When a Unity error is received from REST, we search in all registered
Unity exception and throw the matched one.

:param clz: the exception class
:return: the input exception class
"""

_xmlapi_exception_factory = MappedErrorCodeDecoratorFactory(
    default_exception=VNXBackendError)
get_xmlapi_exception = _xmlapi_exception_factory.get_exception
xmlapi_exception = _xmlapi_exception_factory.clz_decorator()
xmlapi_exception.__doc__ = """ class decorator for VNX File XML API exceptions

Each error from XML API has a unique error code.  We will search all
registered XML API exception and throw the matched one.

:param clz: the exception class
:return: the input exception class
"""

_cli_exception_factory = ExceptionListDecoratorFactory(
    default_exception=VNXException)
get_cli_exception = _cli_exception_factory.get_exception
cli_exception = _cli_exception_factory.clz_decorator()
cli_exception.__doc__ = """ class decorator for VNX CLI exceptions

The decorated exception classes will be used as exception candidate.
When an error message is received in CLI, we will search all CLI
exceptions and throw the matched on.

Each class must either have `error_code` or `error_message` property.

:param clz: the exception class
:return: the input exception class
"""


def _extract_output(output):
    if output is not None and not isinstance(output, six.string_types):
        if hasattr(output, 'message'):
            output = output.message
        elif hasattr(output, 'why'):
            # for EvError
            output = getattr(output, 'why')
        elif hasattr(output, 'hex_problem_message_codes'):
            codes = getattr(output, 'hex_problem_message_codes')
            output = ' '.join(codes)
    return output


def raise_if_err(out, msg=None, default=None):
    out = _extract_output(out)
    ex_clz = get_cli_exception(out, default)

    if msg is None:
        if hasattr(out, 'get_status_msg'):
            msg = out.get_status_msg()
        else:
            msg = out
    else:
        msg = '{}  detail:\n{}'.format(msg, out)

    # check if out is empty
    if out is not None and len(out) > 0:
        raise ex_clz(msg)


def check_error(out, *ex_clz_list):
    """ check whether this cli output contains the specified exception

    :param out: output of naviseccli
    :param ex_clz_list: exception class to check
    :return: nothing, raise `ex_clz` if match
    """
    try:
        raise_if_err(out)
    except ex_clz_list:
        raise
    except StoropsException:
        # swallow other errors
        pass
    return out


def check_nas_cmd_error(output, default=None):
    if default is None:
        default = VNXNasCommandNoError

    try:
        raise_if_err(output, default=default)
    except VNXNasCommandNoError:
        # meaning no error
        pass
    except:  # noqa
        # re-raise the error
        raise


class EnumValueNotFoundError(StoropsException, ValueError):
    pass


class StoropsTestException(Exception):
    pass


class MockFileNotFoundError(StoropsTestException):
    pass


class NoIndexException(StoropsException):
    pass


class UnityPerfMonNotEnabledError(UnityException):
    message = ('Performance metric is not available, because performance '
               'monitoring is not enabled for this unity resource.  '
               'You could enable performance monitoring by executing: '
               'UnitySystem.enable_perf_stats('
               'interval=None, rsc_clz_list=None)')


class UnityNameNotUniqueError(UnityException):
    def __init__(self, message="multiple objects found", objects=None):
        super(UnityNameNotUniqueError, self).__init__()
        self.message = message
        self.objects = objects


class UnityCifsServiceNotEnabledError(UnityException):
    pass


class UnityHostInitiatorNotFoundError(UnityException):
    pass


@rest_exception
class UnityHostInitiatorExistedError(UnityException):
    error_code = 100663539


class UnityHostInitiatorUnknownType(UnityException):
    pass


class UnityAttachAluError(UnityException):
    pass


class UnityAttachError(UnityException):
    pass


class UnityShareShrinkSizeTooLargeError(StoropsException):
    pass


class UnityShareShrinkSizeTooSmallError(StoropsException):
    pass


class UnityHostNameInUseError(UnityException):
    pass


class UnityLunException(UnityException):
    pass


@rest_exception
class UnityPolicyInvalidParametersError(UnityException):
    error_code = 5
    message = 'One or more specified parameters are invalid.'


@rest_exception
class UnityActionNotAllowedError(UnityException):
    error_code = 131149836
    message = 'Action is not allowed for the resource.'


@rest_exception
class UnityLunNameInUseError(UnityLunException):
    error_code = 108007744


@rest_exception
class UnityLunShrinkNotSupportedError(UnityLunException):
    error_code = 108007728


@rest_exception
class UnityLunDeleteSnapshotsAttachedError(UnityLunException):
    error_code = 100666394


@rest_exception
class UnityNothingToModifyError(UnityException):
    error_code = 108007456


@rest_exception
class UnityAttachAluExceedLimitError(UnityAttachAluError):
    error_code = 108008750
    message = "Numbers of LUNs exceeds system limit"


@rest_exception
class UnityAttachExceedLimitError(UnityAttachError):
    error_code = 108008750
    message = "Numbers of LUNs or Snaps exceeds system limit"


class UnityAluAlreadyAttachedError(UnityAttachAluError):
    message = 'Requested LUN has already been added to this host'


class UnityResourceAlreadyAttachedError(UnityAttachError):
    message = 'Requested LUN or Snap has already been attached to this host'


class UnityResourceNotAttachedError(UnityException):
    message = 'Cannot modify the host LUN or Snap which is not attached yet'


@rest_exception
class UnityHluNumberInUseError(UnityException):
    error_code = 100663552


class UnityNoHluAvailableError(UnityException):
    pass


@rest_exception
class UnityResourceNotFoundError(UnityException):
    error_code = 131149829


class UnityResourceNotSupportedError(UnityResourceNotFoundError):
    pass


@rest_exception
class UnityNasServerNameUsedError(UnityException):
    error_code = 108011556


@rest_exception
class UnityNasServerHasFsError(UnityException):
    error_code = 108011564


@rest_exception
class UnitySmbNameInUseError(UnityException):
    error_code = 108011873


@rest_exception
class UnityIpAddressUsedError(UnityException):
    error_code = 108011747


@rest_exception
class UnityOneDnsPerNasServerError(UnityException):
    error_code = 108012064


@rest_exception
class UnitySmbShareNameExistedError(UnityException):
    error_code = 151036420


@rest_exception
class UnityNfsShareNameExistedError(UnityException):
    error_code = 151036164


@rest_exception
class UnityOneSmbServerPerNasServerError(UnityException):
    error_code = 108011888


@rest_exception
class UnityNetBiosNameExistedError(UnityException):
    error_code = 108011876


@rest_exception
class UnityNfsAlreadyEnabledError(UnityException):
    error_code = 108012128


@rest_exception
class UnityFileSystemNameAlreadyExisted(UnityException):
    error_code = 108008464


class UnityCreateSnapError(UnityException):
    pass


class UnitySnapException(UnityException):
    pass


@rest_exception
class UnitySnapNameInUseError(UnitySnapException):
    error_code = (1903001605, 1903132675)


@rest_exception
class UnitySnapAlreadyPromotedException(UnitySnapException):
    error_code = 100666332


@rest_exception
class UnityDeleteAttachedSnapError(UnityException):
    error_code = 1903001603


@rest_exception
class UnityShareOnCkptSnapError(UnityException):
    error_code = 1903001786


@rest_exception
class UnityHostIpInUseError(UnityException):
    error_code = 100663538


@rest_exception
class UnitySmbServerLockedError(UnityException):
    error_code = 108011909


@rest_exception
class UnityAclUserNotFoundError(UnityException):
    error_code = 100663499


@rest_exception
class UnityFileSystemSizeTooSmallError(UnityException):
    error_code = 108008449


class UnityQosException(UnityException):
    pass


@rest_exception
class UnityPolicyNameInUseError(UnityQosException):
    error_code = 151032071


@rest_exception
class UnityPolicyIncorrectRequestBodyError(UnityQosException):
    error_code = 131149877


class UnityEthernetPortMtuSizeNotSupportError(UnityException):
    message = "Specified MTU size is not supported."


class UnityEthernetPortSpeedNotSupportError(UnityException):
    message = "Specified Speed is not supported."


@rest_exception
class UnityVLANUsedByOtherTenantError(UnityException):
    error_code = 118939688


@rest_exception
class UnityTenantUUIDInUse(UnityException):
    error_code = 118939686


@rest_exception
class UnityTenantNameInUseError(UnityException):
    error_code = 118939685


@rest_exception
class UnityVLANAlreadyHasInterfaceError(UnityException):
    error_code = 118939689


class UnityShareTypeNotSupportAccessControlError(UnityException):
    message = 'share type does not support access control.'


class UnityHostNotFoundException(UnityException):
    message = 'specified host not found.'


class UnityConsistencyGroupError(UnityException):
    pass


class UnityConsistencyGroupNameInUseError(UnityConsistencyGroupError):
    message = 'name of the consistency group is already used.'


@rest_exception
class UnityHostAccessAlreadyExistsError(UnityConsistencyGroupError):
    error_code = 108009000


@rest_exception
class UnityStorageResourceNameInUseError(UnityException):
    error_code = 108007952


@rest_exception
class UnityEthernetPortAlreadyAggregatedError(UnityException):
    error_code = 100665643


@rest_exception
class UnityIPInUseException(UnityException):
    error_code = 118939658


class UnityMetricException(UnityException):
    pass


@rest_exception
class UnityMetricQueryNotFoundError(UnityMetricException):
    error_code = 131153932


class SystemAPINotSupported(StoropsException):
    message = "System in version {version} doesn't support this API"


@rest_exception
class UnityHostNotInSameTenantError(StoropsException):
    error_code = 151036194


@rest_exception
class UnityTenantHasNasServerError(StoropsException):
    error_code = 118939691


@rest_exception
class UnityTenantHasHostsError(StoropsException):
    error_code = 118939704


@rest_exception
class UnityPoolNameInUseError(UnityException):
    error_code = 108003600


class VNXStatsError(VNXException):
    pass


class VNXPerMonNotEnabledError(VNXStatsError):
    pass


class NaviseccliNotAvailableError(VNXException):
    message = ("naviseccli not found.  please make sure it's installed"
               " and available in path.")


class VNXNasObjectNotFound(VNXException, VNXObjectNotFoundError):
    message_template = "NAS object is not found.  {err}"


class VNXSetArrayNameError(VNXException):
    pass


class OptionMissingError(VNXException):
    pass


@xmlapi_exception
class VNXInvalidMoverID(VNXException):
    error_code = 14227341323
    message_template = "invalid mover or vdm.  {id}"


class VNXLockRequiredException(VNXException):
    message = "unable to acquire lock."


@cli_exception
class VNXSpNotAvailableError(VNXException):
    error_message = ('connection refused',
                     'A network error occurred while trying to connect')


@cli_exception
class VNXDropConnectionError(VNXException):
    error_message = 'End of data stream'


@cli_exception
class VNXNotSupportedError(VNXException):
    error_message = 'commands are not supported by the target storage system'


class VNXSystemError(VNXException):
    pass


class VNXSystemDownError(VNXSystemError):
    pass


class VNXSPError(VNXException):
    pass


class VNXSPDownError(VNXSPError):
    pass


class VNXStorageGroupError(VNXException):
    pass


class VNXAttachAluError(VNXException):
    pass


@cli_exception
class VNXInvalidCliParamError(VNXException):
    error_message = 'invalid command line parameters'


@cli_exception
class VNXPortNotInitializedError(VNXStorageGroupError):
    error_message = 'port is uninitialized'


@cli_exception
class VNXInitiatorExistedError(VNXStorageGroupError):
    error_message = 'Initiator record already exists'


@cli_exception
class VNXAluAlreadyAttachedError(VNXAttachAluError):
    error_message = (
        'LUN already exists in the specified storage group',
        'Requested LUN has already been added to this Storage Group')


@cli_exception
class VNXAluNotFoundError(VNXAttachAluError, VNXObjectNotFoundError):
    error_message = 'The ALU number specified by user is not a bound'


@cli_exception
class VNXHluNumberInUseError(VNXAttachAluError):
    error_message = 'Requested Host LUN Number already in use'


class VNXDetachAluError(VNXStorageGroupError):
    pass


@cli_exception
class VNXDetachAluNotFoundError(VNXDetachAluError, VNXObjectNotFoundError):
    error_message = 'No such Host LUN in this Storage Group'


class VNXCreateStorageGroupError(VNXStorageGroupError):
    pass


@cli_exception
class VNXStorageGroupNameInUseError(VNXCreateStorageGroupError):
    error_message = 'Storage Group name already in use'


class VNXDeleteStorageGroupError(VNXStorageGroupError):
    pass


@cli_exception
class VNXStorageGroupNotFoundError(VNXStorageGroupError,
                                   VNXObjectNotFoundError):
    error_message = ('The group name or UID does not match any '
                     'storage groups for this array')


class VNXNoHluAvailableError(VNXStorageGroupError):
    pass


class VNXHluAlreadyUsedError(VNXStorageGroupError):
    pass


class VNXMigrationError(VNXException):
    pass


@cli_exception
class VNXLunNotMigratingError(VNXMigrationError):
    error_message = 'The specified source LUN is not currently migrating.'


@cli_exception
class VNXLunSyncCompletedError(VNXMigrationError):
    error_code = 0x714a8021


class VNXIOClassError(VNXException):
    pass


@cli_exception
class VNXIOCLassRunningError(VNXIOClassError):
    error_message = 'Can not modify or delete an I/O Class while it is '
    'currently running'


class VNXIOPolicyError(VNXException):
    pass


@cli_exception
class VNXIOPolicyRunningError(VNXIOPolicyError):
    error_message = 'Can not modify a Policy, delete a Policy or update event '
    'log settings while it is currently running'


@cli_exception
class VNXTargetNotReadyError(VNXMigrationError):
    error_message = 'The destination LUN is not available for migration'


class VNXSnapError(VNXException):
    pass


class VNXModifySnapError(VNXSnapError):
    pass


@cli_exception
class VNXDeleteAttachedSnapError(VNXSnapError):
    error_code = 0x716d8003


@cli_exception
class VNXSnapAlreadyMountedError(VNXSnapError):
    error_code = 0x716d8055


@cli_exception
class VNXSnapNotAttachedError(VNXSnapError):
    error_message = 'Snapshot mount point is not currently attached.'


class VNXCreateSnapError(VNXException):
    pass


class VNXAttachSnapError(VNXSnapError):
    pass


@cli_exception
class VNXAttachSnapLunTypeError(VNXAttachSnapError):
    error_message = 'Cannot attach the snapshot. Invalid LUN type.'


class VNXDetachSnapError(VNXSnapError):
    pass


@cli_exception
class VNXDetachSnapLunTypeError(VNXDetachSnapError):
    error_message = 'it is not a snapshot mount point.'


@cli_exception
class VNXSnapNameInUseError(VNXSnapError):
    error_code = 0x716d8005


@cli_exception
class VNXCreateSnapResourceNotFoundError(VNXSnapError, VNXObjectNotFoundError):
    error_message = 'The specified resource does not exist.'


class VNXDeleteSnapError(VNXSnapError):
    pass


@cli_exception
class VNXSnapNotExistsError(VNXSnapError):
    error_message = 'The specified snapshot does not exist.'


@cli_exception
class VNXRestoreSnapToMismatchLunError(VNXSnapError):
    error_code = 0x716d8008


class VNXLunError(VNXException):
    pass


class VNXCreateLunError(VNXLunError):
    pass


@cli_exception
class VNXLunNameInUseError(VNXCreateLunError):
    error_code = 0x712d8d04


class VNXModifyLunError(VNXLunError):
    pass


class VNXCreateMpError(VNXLunError):
    pass


class VNXLunExtendError(VNXLunError):
    pass


@cli_exception
class VNXNotReadyExpandError(VNXLunExtendError):
    error_code = 0x712d8d33


@cli_exception
class VNXLunExpandSizeError(VNXLunExtendError):
    error_code = 0x712d8e04


@cli_exception
class VNXLunPreparingError(VNXLunError):
    error_code = 0x712d8e0e


@cli_exception
class VNXLunNotFoundError(VNXLunError, VNXObjectNotFoundError):
    error_message = 'Could not retrieve the specified (pool lun).'


class VNXDeleteLunError(VNXLunError):
    pass


@cli_exception
class VNXLunInStorageGroupError(VNXDeleteLunError):
    error_message = ('contained in a Storage Group',
                     'LUN mapping still exists')


@cli_exception
class VNXLunInConsistencyGroupError(VNXDeleteLunError):
    error_code = 0x716d8025


@cli_exception
class VNXLunHasSnapMountPointError(VNXDeleteLunError):
    error_code = 0x716d801a


@cli_exception
class VNXLunHasSnapError(VNXDeleteLunError):
    error_message = ("The LUN cannot be destroyed because it still"
                     " has snapshots")


@cli_exception
class VNXLunUsedByFeatureError(VNXDeleteLunError):
    error_message = ("Cannot unbind LUN because it's being used by a"
                     " feature of the Storage System")


class VNXCompressionError(VNXLunError):
    pass


@cli_exception
class VNXCompressionAlreadyEnabledError(VNXCompressionError):
    error_message = 'Compression on the specified LUN is already turned on.'


class VNXDedupError(VNXLunError):
    pass


@cli_exception
class VNXDedupAlreadyEnabled(VNXDedupError):
    error_message = ['Deduplication is already enabled',
                     'the deduplication state of LUN is enabled or enabling.']


class VNXConsistencyGroupError(VNXException):
    pass


class VNXCreateConsistencyGroupError(VNXConsistencyGroupError):
    pass


@cli_exception
class VNXConsistencyGroupNameInUseError(VNXCreateConsistencyGroupError):
    error_code = 0x716d8021


@cli_exception
class VNXConsistencyGroupNotFoundError(VNXConsistencyGroupError,
                                       VNXObjectNotFoundError):
    error_message = 'Cannot find the consistency group'


@cli_exception
class VNXConsistencyGroupIsDeletingError(VNXConsistencyGroupError):
    error_code = 0x712d8801


class VNXDeleteHbaError(VNXException):
    pass


@cli_exception
class VNXDeleteHbaNotFoundError(VNXException, VNXObjectNotFoundError):
    error_message = 'The HBA UID specified is not known by the array'


class VNXPingNodeError(VNXException):
    pass


@cli_exception
class VNXPingNodeTimeOutError(VNXPingNodeError):
    error_message = 'Request timed out.'


@cli_exception
class VNXPingNodeSuccess(VNXPingNodeError):
    error_regex = r'TTL=\w+'


class VNXSecurityException(VNXException):
    pass


@cli_exception
class VNXCredentialError(VNXSecurityException):
    error_message = ['invalid username, password and/or scope.',
                     'Could not connect to the specified host']


@cli_exception
class VNXUserNameInUseError(VNXSecurityException):
    error_message = 'The specified user already exists.'


@cli_exception
class VNXUserNotFoundError(VNXSecurityException, VNXObjectNotFoundError):
    error_message = 'User does not exist'


class VNXStatsException(VNXException):
    pass


class VNXRaidGroupError(VNXException):
    pass


class VNXCreateRaidGroupError(VNXRaidGroupError):
    pass


class VNXDeleteRaidGroupError(VNXRaidGroupError):
    pass


class VNXPoolError(VNXException):
    pass


class VNXCreatePoolError(VNXPoolError):
    pass


@cli_exception
class VNXNameInUseError(VNXException):
    error_message = 'Could not set properties:(Name).'


@cli_exception
class VNXPoolNameInUseError(VNXCreatePoolError, VNXNameInUseError):
    error_message = ['0x712d8501', 'Pool name is already used']


@cli_exception
class VNXDiskUsedError(VNXCreatePoolError):
    error_code = 0x76008304


class VNXModifyPoolError(VNXPoolError):
    pass


class VNXDeletePoolError(VNXPoolError):
    pass


@cli_exception
class VNXPoolDestroyingError(VNXDeletePoolError):
    error_message = 'the Storage Pool because it is Destroying'


@cli_exception
class VNXPoolNotFoundError(VNXPoolError, VNXObjectNotFoundError):
    error_message = ['The (Storagepool) may not exist',
                     'was not found in any provider']


class VNXNotEnoughDiskAvailableError(VNXPoolError):
    message = 'not enough disk for pool creation.'


class VNXMirrorException(VNXException):
    pass


@cli_exception
class VNXMirrorLunNotAvailableError(VNXMirrorException):
    error_message = 'Specified LU not available for mirroring.'


@cli_exception
class VNXMirrorNameInUseError(VNXMirrorException):
    error_message = 'Mirror name already in use'


@cli_exception
class VNXMirrorAlreadyMirroredError(VNXMirrorException):
    error_message = ('A mirror image for the specified mirror '
                     'already exists on the chosen subsystem')


@cli_exception
class VNXMirrorImageNotFoundError(VNXMirrorException, VNXObjectNotFoundError):
    error_message = 'Image not found'


@cli_exception
class VNXMirrorFractureImageError(VNXMirrorException):
    error_message = 'Cannot fracture the image because of'


@cli_exception
class VNXMirrorSyncImageError(VNXMirrorException):
    error_message = 'Synchronizing sync mirror image failed'


@cli_exception
class VNXMirrorPromoteNonLocalImageError(VNXMirrorException):
    error_code = 0x7105824e


@cli_exception
class VNXMirrorRemoveSynchronizingError(VNXMirrorException):
    error_code = 0x71058247


@cli_exception
class VNXMirrorPromotePrimaryError(VNXMirrorException):
    error_message = 'Cannot remove or promote a primary image.'


@cli_exception
class VNXMirrorFeatureNotAvailableError(VNXMirrorException):
    error_message = 'Mirror Feature software is not installed'


@cli_exception
class VNXMirrorNotFoundError(VNXMirrorException, VNXObjectNotFoundError):
    error_message = 'Mirror not found'


@cli_exception
class VNXDeleteMirrorWithSecondaryError(VNXMirrorException):
    error_code = 0x71058243


@cli_exception
class VNXMirrorGroupNameInUseError(VNXMirrorException):
    error_message = 'Mirror Group name already in use'


@cli_exception
class VNXMirrorGroupNotFoundError(VNXMirrorException, VNXObjectNotFoundError):
    error_message = 'Unable to locate the specified group'


@cli_exception
class VNXMirrorGroupAlreadyMemberError(VNXMirrorException):
    error_message = 'The mirror is already a member of a group'


@cli_exception
class VNXMirrorGroupMirrorNotMemberError(VNXMirrorException):
    error_message = 'The specified mirror is not a member of the group'


@cli_exception
class VNXMirrorGroupAlreadyPromotedError(VNXMirrorException):
    error_message = 'The Consistency Group has no secondary images to promote'


class VNXPortError(VNXException):
    pass


@cli_exception
class VNXGateWayError(VNXException):
    error_message = ('The gateway is not on the subnet defined '
                     'by the IP address and netmask.')


@cli_exception
class VNXVirtualPortNotFoundError(VNXPortError, VNXObjectNotFoundError):
    error_message = 'Request failed. Specified virtual port not found.'


@xmlapi_exception
class VNXGeneralNasError(VNXException):
    error_code = 13690601492


class VNXVdmError(VNXException):
    pass


@xmlapi_exception
class VNXInvalidVdmIdError(VNXVdmError):
    error_code = 13421840550


@xmlapi_exception
class VNXVdmAlreadyExistedError(VNXVdmError):
    error_code = 13421840550


class VNXFsError(VNXException):
    pass


@xmlapi_exception
class VNXFsExistedError(VNXFsError):
    error_code = 13691191325


@xmlapi_exception
class VNXFsNotFoundError(VNXFsError, VNXObjectNotFoundError):
    error_code = 18522112101


class VNXFsSnapError(VNXException):
    pass


@xmlapi_exception
class VNXFsSnapNameInUseError(VNXFsSnapError):
    error_code = 13690535947


class VNXMoverInterfaceError(VNXException):
    pass


@xmlapi_exception
class VNXMoverInterfaceNotFoundError(VNXMoverInterfaceError,
                                     VNXObjectNotFoundError):
    error_code = 13691781134


@cli_exception
class VNXMoverInterfaceNotAttachedError(VNXMoverInterfaceError):
    error_message = 'is not currently attached to the VDM'


@xmlapi_exception
class VNXMoverInterfaceInvalidVlanIdError(VNXMoverInterfaceError):
    error_code = 13421850371


@xmlapi_exception
class VNXMoverInterfaceExistedError(VNXMoverInterfaceError):
    error_code = 13691781136


@xmlapi_exception
class VNXMoverInterfaceNameInUseError(VNXMoverInterfaceError):
    error_code = 13421840550


class VNXFileCredentialError(VNXCredentialError):
    message = 'credential error for VNX control station.'


class VNXNasCommandNoError(VNXException):
    """ Nas command returns something even if command success.

    Use this error as the default exception for output verification.
    If this exception is thrown, we know that no error is found.
    """
    pass


@cli_exception
class VNXMoverInterfaceNotExistsError(VNXMoverInterfaceError):
    error_regex = 'network interface .* does not exist'


class UnityJobException(UnityException):
    """Unity Job exception.

    Any job related exception should inherit this exception."""
    pass


class JobTimeoutException(UnityJobException):
    message = "Timeout when waiting for job completion."


class JobStateError(UnityJobException):
    def __init__(self, job):
        self.job = job
        super(JobStateError, self).__init__(self.message)

    @property
    def message(self):
        return 'Job State: {}.  Error Detail: {}'.format(
            self.job.state.name, '.  '.join(self.job.messages))


class UnityThinCloneException(UnityException):
    """Unity exceptions for Thin clone.

    Any thin-clone related exceptions should inherit this exception."""
    pass


@rest_exception
class UnityThinCloneLimitExceededError(UnityThinCloneException):
    error_code = 108008767


@rest_exception
class UnityBaseHasThinCloneError(UnityThinCloneException):
    error_code = 108009075


@rest_exception
class UnityTCSnapUnderDestroyError(UnityThinCloneException):
    error_code = 108008719


class UnityCGMemberActionNotSupportError(UnityException):
    message = ('The action not support on lun in cg or member snap of cg snap.'
               ' Execute it on cg level.')


class UnityThinCloneNotAllowedError(UnityException):
    message = 'Thinclone not support on thick luns or snaps of thick lun.'


class UnityMigrationException(UnityException):
    """Unity exceptions for Migration.

    Any migration related exceptions should inherit this exception."""
    pass


@rest_exception
class UnityMigrationSourceDestNotExistsError(UnityMigrationException):
    error_code = 151036683


@rest_exception
class UnityMigrationSourceIsThinCloneError(UnityMigrationException):
    error_code = 151036719


@rest_exception
class UnityMigrationSourceHasThinCloneError(UnityMigrationException):
    error_code = 151036720


class UnityMigrationTimeoutException(UnityMigrationException):
    message = "Timeout when waiting for lun migration."


@rest_exception
class UnityDeleteLunInReplicationError(UnityException):
    error_code = 105906485


class UnityImportSessionException(UnityException):
    """Unity exceptions for Import Session.

    Any import session related exceptions should inherit this exception."""
    pass


@rest_exception
class UnityVNXSANCopyHostNotExistsError(UnityImportSessionException):
    error_code = 105906713


@rest_exception
class UnitySourceResourceInAnImportSessionError(UnityImportSessionException):
    error_code = 105906723


class StoropsConnectTimeoutError(StoropsException):
    pass


class UnityReplicationError(UnityException):
    pass


class UnityLocalReplicationFsNameNotSameError(UnityReplicationError):
    pass


@rest_exception
class UnityFileResourceReplicationInUseError(UnityReplicationError):
    error_code = 105906504


@rest_exception
class UnityDeleteShareFromDestNasServerError(UnityReplicationError):
    error_code = 151036197


@rest_exception
class UnityFailoverNasRepWithFsRepOutOfSyncError(UnityReplicationError):
    error_code = 105906786


class UnitySnapScheduleNameInUseError(UnityException):
    pass



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import random
from itertools import chain

import six

from retryz import retry

from storops import exception as ex
from storops.lib import common
from storops.lib import converter
from storops.lib.version import version
from storops.unity.enums import HostTypeEnum, HostInitiatorTypeEnum
from storops.unity.resource import UnityResource, UnityResourceList, \
    UnityAttributeResource
from storops.unity.resource.tenant import UnityTenant

# from storops.unity.resource import lun

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class UnityBlockHostAccess(UnityAttributeResource):
    pass


class UnityBlockHostAccessList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityBlockHostAccess

    def get_host_id(self):
        return [access.host.id for access in self]


class UnitySnapHostAccess(UnityAttributeResource):
    pass


class UnitySnapHostAccessList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnitySnapHostAccess


DUMMY_LUN_NAME = 'storops_dummy_lun'
MAX_HLU_NUMBER = 255  # Unity supports 16381 but uses 255 here due to GH-238


class UnityHost(UnityResource):
    @classmethod
    def get_nested_properties(cls):
        return (
            'fc_host_initiators.initiator_id',
            'fc_host_initiators.paths.initiator.type',
            'fc_host_initiators.paths.is_logged_in',
            'fc_host_initiators.paths.fc_port.wwn',
            'iscsi_host_initiators.initiator_id',
            'hostLUNs.hlu',
            'hostLUNs.lun.id',
            'hostLUNs.lun.name',
            'hostLUNs.snap.id',
            'hostIPPorts.address',
        )

    @classmethod
    def create(cls, cli, name, host_type=None, desc=None, os=None,
               tenant=None):
        if host_type is None:
            host_type = HostTypeEnum.HOST_MANUAL

        if tenant is not None:
            tenant = UnityTenant.get(cli, tenant)

        resp = cli.post(cls().resource_class,
                        type=host_type,
                        name=name,
                        description=desc,
                        osType=os,
                        tenant=tenant)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    @classmethod
    def get_host(cls, cli, _id, force_create=False, tenant=None):
        if isinstance(_id, six.string_types) and ('.' in _id or ':' in _id):
            # it looks like an ip address, find or create the host
            address, prefix, netmask = converter.parse_host_address(_id)
            ports = UnityHostIpPortList(cli=cli, address=address)
            # since tenant is not supported by all kinds of system. So we
            # should avoid send the tenant request if tenant is None
            tenant = None if tenant is None else UnityTenant.get(cli, tenant)
            ports = [port for port in ports if port.host.tenant == tenant]

            if len(ports) == 1:
                ret = ports[0].host
            elif force_create:
                log.info('cannot find an existing host with ip {}.  '
                         'create a new host "{}" to attach it.'
                         .format(address, address))
                host_type = (HostTypeEnum.SUBNET if netmask
                             else HostTypeEnum.HOST_MANUAL)
                host_name = ('{}_{}'.format(address, netmask) if netmask
                             else address)
                host = cls.create(cli, host_name, host_type=host_type,
                                  tenant=tenant)
                if ':' in address:  # ipv6
                    host.add_ip_port(address, v6_prefix_length=prefix)
                else:  # ipv4
                    host.add_ip_port(address, netmask=netmask)
                ret = host.update()
            else:
                ret = None
        else:
            ret = cls.get(cli=cli, _id=_id)
        return ret

    def modify(self, name=None, desc=None, os=None):
        req_body = self._cli.make_body(
            name=name,
            description=desc,
            osType=os,
        )
        resp = self._cli.modify(self.resource_class,
                                self.get_id(), **req_body)
        resp.raise_if_err()
        return self

    def modify_host_lun(self, lun_or_snap, new_hlu):
        host_lun = self.get_host_lun(lun_or_snap)
        if not host_lun:
            raise ex.UnityResourceNotAttachedError
        self._modify_hlu(host_lun, new_hlu)

    def _modify_hlu(self, host_lun, new_hlu):
        resp = self._cli.action(self.resource_class, self.get_id(),
                                'modifyHostLUNs',
                                hostLunModifyList=[{'hostLUN': host_lun,
                                                    'hlu': new_hlu}])
        resp.raise_if_err()

    def _get_host_luns(self, lun=None, snap=None, hlu=None):
        ret = self.host_luns

        if ret is None:
            log.debug('No lun attached to the host: {}.'.format(self.name))
            return []
        else:
            lun_id = lun.id if lun is not None else None
            snap_id = snap.id if snap is not None else None
            hlu_no = hlu if hlu is not None else None

            if lun_id is not None:
                ret = filter(lambda x: x.lun and x.lun.get_id() == lun_id, ret)

            if snap_id is None:
                # get rid of the `Snapshot` type of host access
                ret = filter(lambda x: x.snap is None, ret)
            else:
                ret = filter(lambda x: x.snap and x.snap.get_id() == snap_id,
                             ret)

            if hlu_no is not None:
                ret = filter(lambda x: x.hlu == hlu_no, ret)

            ret = list(ret)
            msg = ('Found {num} host luns attached to this host. '
                   'Filter: lun={lun_id}, snap={snap_id}, '
                   'hlu={hlu_no}.').format(num=len(ret), lun_id=lun_id,
                                           snap_id=snap_id, hlu_no=hlu_no)
            log.debug(msg)

            return ret

    def detach(self, lun_or_snap):
        if self.host_luns:
            # To detach the `dummy luns` which are attached via legacy storops.
            dummy_lun_ids = [lun.get_id() for lun in self.host_luns.lun
                             if lun.name == DUMMY_LUN_NAME]
            if dummy_lun_ids:
                from storops.unity.resource.lun import UnityLun
                dummy_lun = UnityLun(cli=self._cli, _id=dummy_lun_ids[0])
                try:
                    dummy_lun.delete(None)
                except ex.UnityException:
                    pass
        return lun_or_snap.detach_from(self)

    def detach_alu(self, lun):
        log.warn('Method detach_alu is deprecated. Use detach instead.')
        return lun.detach_from(self)

    def _random_hlu_number(self):
        existing_hlus = [host_lun.hlu for host_lun in self._get_host_luns()]
        try:
            candidate = random.choice(
                list(hlu for hlu in range(1, MAX_HLU_NUMBER + 1)
                     if hlu not in existing_hlus))
            log.debug('Random choose hlu: {hlu} from range [1, {max}].'.format(
                hlu=candidate, max=MAX_HLU_NUMBER))
            return candidate
        except IndexError:
            raise ex.UnityNoHluAvailableError(
                'No hlu available. Random choose hlu from range [1, {max}] '
                'but not in existing hlu: {exist}'.format(exist=existing_hlus,
                                                          max=MAX_HLU_NUMBER))

    @version('<4.4.0')  # noqa
    @retry(limit=5, on_error=ex.UnityHluNumberInUseError)
    def _attach_with_retry(self, lun_or_snap, skip_hlu_0):
        # Before 4.4.0 (Osprey), it didn't support to pass in hlu when
        # attaching.
        lun_or_snap.attach_to(self)
        self.update()
        host_lun = self.get_host_lun(lun_or_snap)
        if skip_hlu_0 and host_lun.hlu == 0:
            candidate_hlu = self._random_hlu_number()
            self._modify_hlu(host_lun, candidate_hlu)
            return candidate_hlu
        else:
            return host_lun.hlu

    @version('>=4.4.0')  # noqa
    @retry(limit=5, on_error=ex.UnityHluNumberInUseError)
    def _attach_with_retry(self, lun_or_snap, skip_hlu_0):
        # From 4.4.0 (Osprey), it supported to pass in hlu when attaching LUN,
        # but not when attaching snap.
        from storops.unity.resource.lun import UnityLun as _UnityLun
        if skip_hlu_0 and isinstance(lun_or_snap, _UnityLun):
            candidate_hlu = self._random_hlu_number()
            lun_or_snap.attach_to(self, hlu=candidate_hlu)
            self.update()
            return candidate_hlu
        else:
            lun_or_snap.attach_to(self)
            self.update()
            host_lun = self.get_host_lun(lun_or_snap)
            return host_lun.hlu

    def attach(self, lun_or_snap, skip_hlu_0=False):
        """ Attaches lun, snap or member snap of cg snap to host.

        Don't pass cg snapshot in as `lun_or_snap`.

        :param lun_or_snap: the lun, snap, or a member snap of cg snap
        :param skip_hlu_0: whether to skip hlu 0
        :return: the hlu number
        """

        # `UnityResourceAlreadyAttachedError` check was removed due to there
        # is a host cache existing in Cinder driver. If the lun was attached to
        # the host and the info was stored in the cache, wrong hlu would be
        # returned.
        # And attaching a lun to a host twice would success, if Cinder retry
        # triggers another attachment of same lun to the host, the cost would
        # be one more rest request of `modifyLun` and one for host instance
        # query.
        try:
            return self._attach_with_retry(lun_or_snap, skip_hlu_0)

        except ex.SystemAPINotSupported:
            # Attaching snap to host not support before 4.1.
            raise
        except ex.UnityAttachExceedLimitError:
            # The number of luns exceeds system limit
            raise
        except:  # noqa
            # other attach error, remove this lun if already attached
            self.detach(lun_or_snap)
            raise

    def attach_alu(self, lun):
        log.warn('Method attach_alu is deprecated. Use attach instead.')
        if self.has_alu(lun):
            raise ex.UnityAluAlreadyAttachedError()

        try:
            lun.attach_to(self)
            self.update()
            hlu = self.get_hlu(lun)
        except ex.UnityAttachAluExceedLimitError:
            # The number of luns exceeds system limit
            raise
        except:  # noqa
            # other attach error, remove this lun if already attached
            self.detach_alu(lun)
            raise

        return hlu

    def has_hlu(self, lun_or_snap, cg_member=None):
        """Returns True if `lun_or_snap` is attached to the host.

        :param lun_or_snap: can be lun, lun snap, cg snap or a member snap of
            cg snap.
        :param cg_member: the member lun of cg if `lun_or_snap` is cg snap.
        :return: True - if `lun_or_snap` is attached, otherwise False.
        """
        hlu = self.get_hlu(lun_or_snap, cg_member=cg_member)
        return hlu is not None

    def has_alu(self, lun):
        log.warn('Method has_alu is deprecated. Use has_hlu instead.')
        alu = self.get_hlu(lun)
        if alu is None:
            return False
        else:
            return True

    def get_host_lun(self, lun_or_snap, cg_member=None):
        """Gets the host lun of a lun, lun snap, cg snap or a member snap of cg
        snap.

        :param lun_or_snap: can be lun, lun snap, cg snap or a member snap of
            cg snap.
        :param cg_member: the member lun of cg if `lun_or_snap` is cg snap.
        :return: the host lun object.
        """
        import storops.unity.resource.lun as lun_module
        import storops.unity.resource.snap as snap_module
        which = None
        if isinstance(lun_or_snap, lun_module.UnityLun):
            which = self._get_host_luns(lun=lun_or_snap)
        elif isinstance(lun_or_snap, snap_module.UnitySnap):
            if lun_or_snap.is_cg_snap():
                if cg_member is None:
                    log.debug('None host lun for CG snap {}. '
                              'Use its member snap instead or pass in '
                              'cg_member.'.format(lun_or_snap.id))
                    return None
                lun_or_snap = lun_or_snap.get_member_snap(cg_member)
                which = self._get_host_luns(lun=cg_member, snap=lun_or_snap)
            else:
                which = self._get_host_luns(snap=lun_or_snap)
        if not which:
            log.debug('Resource(LUN or Snap) {} is not attached to host {}'
                      .format(lun_or_snap.name, self.name))
            return None
        return which[0]

    def get_hlu(self, resource, cg_member=None):
        """Gets the hlu number of a lun, lun snap, cg snap or a member snap of
        cg snap.

        :param resource: can be lun, lun snap, cg snap or a member snap of cg
            snap.
        :param cg_member: the member lun of cg if `lun_or_snap` is cg snap.
        :return: the hlu number.
        """
        host_lun = self.get_host_lun(resource, cg_member=cg_member)
        return host_lun if host_lun is None else host_lun.hlu

    def add_initiator(self, uid, force_create=True, **kwargs):
        initiators = UnityHostInitiatorList.get(cli=self._cli,
                                                initiator_id=uid)

        if not initiators:
            # Set the ISCSI or FC type
            if common.is_fc_uid(uid):
                uid_type = HostInitiatorTypeEnum.FC
            elif common.is_iscsi_uid(uid):
                uid_type = HostInitiatorTypeEnum.ISCSI
            else:
                uid_type = HostInitiatorTypeEnum.UNKNOWN

            if force_create:
                initiator = UnityHostInitiator.create(self._cli, uid,
                                                      self, uid_type, **kwargs)
            else:
                raise ex.UnityHostInitiatorNotFoundError(
                    'name {} not found under host {}.'.format(uid, self.name))
        else:
            initiator = initiators.first_item
            log.debug('initiator {} is existed in unity system.'.format(uid))

        initiator.modify(self)
        return initiator.update()

    def delete_initiator(self, uid):
        initiators = []
        if self.fc_host_initiators:
            initiators += self.fc_host_initiators
        if self.iscsi_host_initiators:
            initiators += self.iscsi_host_initiators
        for item in initiators:
            if item.initiator_id == uid:
                resp = item.delete()
                resp.raise_if_err()
                break
        else:
            raise ex.UnityHostInitiatorNotFoundError(
                'name {} not found under host {}.'.format(uid, self.name))

        return resp

    def add_ip_port(self, address, netmask=None, v6_prefix_length=None,
                    is_ignored=None):
        return UnityHostIpPort.create(self._cli,
                                      host=self,
                                      address=address,
                                      netmask=netmask,
                                      v6_prefix_length=v6_prefix_length,
                                      is_ignored=is_ignored)

    def delete_ip_port(self, address):
        for ip_port in self.host_ip_ports:
            if ip_port.address == address:
                resp = ip_port.delete()
                break
        else:
            resp = None
            log.info('ip {} not found under host {}.'
                     .format(address, self.name))
        return resp

    def update_initiators(self, iqns=None, wwns=None):
        """Primarily for puppet-unity use.

        Update the iSCSI and FC initiators if needed.
        """
        # First get current iqns
        iqns = set(iqns) if iqns else set()
        current_iqns = set()
        if self.iscsi_host_initiators:
            current_iqns = {initiator.initiator_id
                            for initiator in self.iscsi_host_initiators}
        # Then get current wwns
        wwns = set(wwns) if wwns else set()
        current_wwns = set()
        if self.fc_host_initiators:
            current_wwns = {initiator.initiator_id
                            for initiator in self.fc_host_initiators}
        updater = UnityHostInitiatorUpdater(
            self, current_iqns | current_wwns, iqns | wwns)
        return updater.update()

    @property
    def ip_list(self):
        if self.host_ip_ports:
            ret = [port.address for port in self.host_ip_ports]
        else:
            ret = []
        return ret


class UnityHostInitiatorUpdater(object):
    def __init__(self, host, current, new_initiators):
        self.host = host
        self.current = current
        self.new_initiators = new_initiators

    def compute(self):
        to_add = self.new_initiators - self.current
        to_delete = self.current - self.new_initiators
        return to_add, to_delete

    def update(self):
        to_add, to_delete = self.compute()
        for a in to_add:
            self.host.add_initiator(a)
        for d in to_delete:
            self.host.delete_initiator(d)
        return len(to_add | to_delete)


class UnityHostList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityHost

    @property
    def ip_list(self):
        return list(chain.from_iterable([host.ip_list for host in self]))


class UnityHostInitiator(UnityResource):
    @classmethod
    def create(cls, cli, uid, host, type, is_ignored=None,
               chap_user=None, chap_secret=None, chap_secret_type=None):

        if type == HostInitiatorTypeEnum.ISCSI:
            resp = cli.post(cls().resource_class,
                            host=host,
                            initiatorType=type,
                            initiatorWWNorIqn=uid,
                            chapUser=chap_user,
                            chapSecret=chap_secret,
                            chapSecretType=chap_secret_type,
                            isIgnored=is_ignored)
        elif type == HostInitiatorTypeEnum.FC:
            resp = cli.post(cls().resource_class,
                            host=host,
                            initiatorType=type,
                            initiatorWWNorIqn=uid,
                            isIgnored=is_ignored)
        else:
            raise ex.UnityHostInitiatorUnknownType(
                '{} parameter is unknown type'.format(type))

        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    def modify(self, host, is_ignored=None, chap_user=None,
               chap_secret=None, chap_secret_type=None):
        req_body = {'host': host, 'isIgnored': is_ignored}

        if self.type == HostInitiatorTypeEnum.ISCSI:
            req_body['chapUser'] = chap_user
            req_body['chapSecret'] = chap_secret
            req_body['chapSecretType'] = chap_secret_type
        # end if

        resp = self._cli.modify(self.resource_class,
                                self.get_id(), **req_body)
        resp.raise_if_err()
        return resp


class UnityHostInitiatorList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityHostInitiator


class UnityHostInitiatorPath(UnityResource):
    pass


class UnityHostInitiatorPathList(UnityResourceList):
    def __init__(self, cli=None, is_logged_in=None, **filters):
        super(UnityHostInitiatorPathList, self).__init__(cli, **filters)
        self._is_logged_in = None
        self._set_filter(is_logged_in)

    def _set_filter(self, is_logged_in=None, **kwargs):
        self._is_logged_in = is_logged_in

    def _filter(self, initiator_path):
        ret = True
        if self._is_logged_in is not None:
            ret &= (initiator_path.initiator.type == HostInitiatorTypeEnum.FC
                    and initiator_path.is_logged_in == self._is_logged_in)
        return ret

    @classmethod
    def get_resource_class(cls):
        return UnityHostInitiatorPath


class UnityHostIpPort(UnityResource):
    @classmethod
    def create(cls, cli, host, address, netmask=None, v6_prefix_length=None,
               is_ignored=None):
        host = UnityHost.get(cli=cli, _id=host)

        resp = cli.post(cls().resource_class,
                        host=host,
                        address=address,
                        netmask=netmask,
                        v6PrefixLength=v6_prefix_length,
                        isIgnored=is_ignored)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)


class UnityHostIpPortList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityHostIpPort


class UnityHostLun(UnityResource):
    pass


class UnityHostLunList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityHostLun



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.exception import raise_if_err, VNXPingNodeError, \
    VNXPingNodeSuccess, VNXPortError
from storops.lib.common import check_int, instance_cache
from storops.vnx.enums import VNXSPEnum, VNXPortType
from storops.vnx.resource import VNXCliResourceList, VNXCliResource

__author__ = 'Cedric Zhuang'


class VNXPort(VNXCliResource):
    """ Base class for all kinds of ports

    Include the basic property like sp, port_id and vport_id.
    """
    @property
    def vport_id(self):
        return self._get_property('_vport_id')

    @property
    @instance_cache
    def index(self):
        sp = 'A' if self.sp == VNXSPEnum.SP_A else 'B'
        if self.vport_id is not None:
            ret = '{}_{}_{}'.format(sp, self.port_id, self.vport_id)
        else:
            ret = '{}_{}'.format(sp, self.port_id)
        return ret

    @property
    def display_name(self):
        return self.index.replace('_', '-')

    @classmethod
    def delete_hba(cls, cli, hba_uid):
        out = cli.delete_hba(hba_uid)
        raise_if_err(out)

    def _get_property(self, prop_name):
        ret = None
        if hasattr(self, prop_name):
            ret = getattr(self, prop_name)

        if ret is None:
            try:
                striped_name = prop_name.strip('_')
                ret = self._get_property_from_raw(striped_name)
            except AttributeError:
                pass
        return ret

    def __hash__(self):
        return hash('<VNXPort {{sp: {}, port_id: {}}}'
                    .format(self.sp, self.port_id))

    @staticmethod
    def _get_inst_prop(instance, name):
        if hasattr(instance, name):
            ret = getattr(instance, name)
        else:
            ret = None
        return ret

    def __eq__(self, other):
        o_sp = self._get_inst_prop(other, 'sp')
        o_port_id = self._get_inst_prop(other, 'port_id')
        o_vport_id = self._get_inst_prop(other, 'vport_id')

        ret = True
        ret &= self.sp == o_sp
        ret &= self.port_id == o_port_id
        ret &= self.vport_id == o_vport_id

        return ret

    def ping_node(self, address, packet_size=None, count=None, timeout=None,
                  delay=None):
        out = self._cli.ping_node(address=address,
                                  sp=self.sp,
                                  port_id=self.port_id,
                                  vport_id=self.virtual_port_id,
                                  packet_size=packet_size,
                                  count=count,
                                  timeout=timeout,
                                  delay=delay)
        try:
            raise_if_err(out, default=VNXPingNodeError)
        except VNXPingNodeSuccess:
            # ping success, pass
            pass

    def config_ip(self, ip, mask, gateway, vport_id=None, vlan_id=None):
        if self.type != VNXPortType.ISCSI:
            raise TypeError('configure IP only works for iSCSI ports.')
        if vport_id is None:
            vport_id = self.vport_id

        out = self._cli.config_iscsi_ip(
            self.sp, self.port_id, ip, mask, gateway, vport_id=vport_id,
            vlan_id=vlan_id)
        raise_if_err(out, default=VNXPortError)

        if vport_id is None:
            vport_id = 0
        return VNXConnectionPort(self.sp, self.port_id, vport_id, self._cli)

    def delete_ip(self, vport_id=None):
        if self.type != VNXPortType.ISCSI:
            raise TypeError('delete IP only works for iSCSI ports.')
        if vport_id is None:
            vport_id = self.vport_id

        out = self._cli.delete_iscsi_ip(self.sp, self.port_id, vport_id)
        raise_if_err(out, default=VNXPortError)


class VNXSPPortList(VNXCliResourceList):
    def __init__(self, cli, sp=None, port_id=None, port_type=None):
        super(VNXSPPortList, self).__init__(cli=cli)
        self._sp = sp
        self._port_id = port_id
        self._port_type = port_type

    def _filter(self, item):
        ret = True

        if self._sp is not None:
            ret &= item.sp == self._sp

        if self._port_id is not None:
            ret &= item.port_id == self._port_id

        if self._port_type is not None:
            ret &= item.type == self._port_type

        return ret

    @classmethod
    def get_resource_class(cls):
        return VNXSPPort

    def _get_raw_resource(self):
        return self._cli.get_sp_port(poll=self.poll)

    @property
    @instance_cache
    def _port_index_map(self):
        return {port.index: port for port in self}

    def get(self, index):
        if isinstance(index, VNXSPPort):
            index = index.index

        return self._port_index_map.get(index)


class VNXSPPort(VNXPort):
    def __init__(self, sp=None, port_id=None, cli=None):
        super(VNXSPPort, self).__init__()
        self._cli = cli
        self._sp = sp
        self._port_id = port_id

    def _get_raw_resource(self):
        raise ValueError('Cannot get single sp port info from cli.  '
                         'Use {}.get(sp, port_id, cli) instead.'
                         .format(self.__class__.__name__))

    @classmethod
    def get(cls, cli, sp=None, port_id=None, port_type=None):
        return VNXSPPortList(cli, sp=sp, port_id=port_id, port_type=port_type)

    @property
    def type(self):
        return VNXPortType.parse(self.wwn)

    def property_names(self):
        ret = super(VNXSPPort, self).property_names()
        ret.append('type')
        return ret


class VNXHbaPort(VNXPort):
    @classmethod
    def _get_parser(cls):
        raise ValueError('property not found.')

    def __init__(self, sp, port_id, vport_id=None):
        super(VNXHbaPort, self).__init__()
        self._sp = VNXSPEnum.parse(sp)
        self._port_id = check_int(port_id)
        self._vport_id = check_int(vport_id, allow_none=True)
        self._type = VNXPortType.FC
        self._host_initiator_list = []

    def is_valid(self):
        return self.sp is not None and self.port_id is not None

    @property
    def sp(self):
        return self._sp

    def get_sp_index(self):
        return VNXSPEnum.get_sp_index(self.sp)

    @property
    def port_id(self):
        return self._port_id

    @property
    def vport_id(self):
        return self._vport_id

    @property
    def type(self):
        ret = None
        if self.host_initiator_list:
            ret = VNXPortType.parse(self.host_initiator_list[0])
        return ret

    @property
    def host_initiator_list(self):
        return tuple(self._host_initiator_list)

    @staticmethod
    def create(sp, port_id, port_type=VNXPortType.FC, vport_id=None):
        port = VNXHbaPort(sp, port_id, vport_id)
        port._type = port_type
        return port

    @staticmethod
    def from_storage_group_hba(sg_hba):
        port = VNXHbaPort.create(sg_hba.sp, sg_hba.port_id,
                                 vport_id=sg_hba.vport_id)
        port._host_initiator_list.append(sg_hba.hba[0])
        return port

    def as_tuple(self):
        return self.sp, self.port_id

    def get_index(self):
        return 'sp'

    def property_names(self):
        return ['sp', 'port_id', 'vport_id', 'type',
                'host_initiator_list']

    def __hash__(self):
        if self.vport_id is not None:
            return hash('<VNXPort {{sp: {}, port_id: {}, vport_id: {}}}'
                        .format(self.sp, self.port_id, self.vport_id))
        else:
            return super(VNXHbaPort, self).__hash__()


class VNXConnectionPortList(VNXCliResourceList):
    def __init__(self, cli, sp=None, port_id=None, vport_id=None,
                 port_type=None, has_ip=None):
        super(VNXConnectionPortList, self).__init__(cli=cli)
        self._sp = sp
        self._port_id = port_id
        self._vport_id = vport_id
        self._port_type = port_type
        self._has_ip = has_ip

    def _filter(self, item):
        ret = True
        if self._sp is not None:
            ret &= item.sp == self._sp

        if self._port_id is not None:
            ret &= item.port_id == self._port_id

        if self._vport_id is not None:
            ret &= item.virtual_port_id == self._vport_id

        if self._port_type is not None:
            ret &= item.type == self._port_type

        if self._has_ip is not None:
            if self._has_ip:
                ret &= item.ip_address is not None
            else:
                ret &= item.ip_address is None
        return ret

    @classmethod
    def get_resource_class(cls):
        return VNXConnectionPort

    def _get_raw_resource(self):
        return self._cli.get_connection_port(poll=self.poll)

    def __len__(self):
        return len(self._virtual_port_list())

    def __iter__(self):
        return iter(self._virtual_port_list())

    def __getitem__(self, item):
        return self._virtual_port_list()[item]

    def _virtual_port_list(self):
        temp_list = super(VNXConnectionPortList, self).list
        vport_lists = []
        for item in temp_list:
            if len(item.virtual_ports):
                ports = self._flat_vports(item)
                vport_lists.extend(ports)
            else:
                vport = VNXConnectionVirtualPort(
                    cli=self._cli, sp=item.sp, port_id=item.port_id,
                    vport_id=item.vport_id)
                self._set_child_props(item, vport)
                vport_lists.append(vport)
        return vport_lists

    def _flat_vports(self, connection_port):
        """Flat the virtual ports."""
        vports = []
        for vport in connection_port.virtual_ports:
            self._set_child_props(connection_port, vport)
            vports.append(vport)
        return vports

    def _set_child_props(self, parent, child):
        props = parent.property_names()
        # Ignore virtual_ports for child
        props = (x for x in props if x != 'virtual_ports')
        for prop_name in props:
            setattr(child, prop_name,
                    getattr(parent, prop_name))

    def get_dict_repr(self, dec=1):
        items = [item.get_dict_repr(dec - 1) for
                 item in self._virtual_port_list()]
        return {self.__class__.__name__: items}


class VNXConnectionPort(VNXPort):
    def __init__(self, sp=None, port_id=None, vport_id=None, cli=None):
        super(VNXConnectionPort, self).__init__()
        self._sp = sp
        self._port_id = port_id
        self._vport_id = vport_id
        self._cli = cli

    @property
    def type(self):
        ret = VNXPortType.parse(self.wwn)
        if ret == VNXPortType.FC and self.virtual_port_id is not None:
            ret = VNXPortType.FCOE
        return ret

    @property
    def existed(self):
        return self.wwn is not None

    def _get_raw_resource(self):
        return self._cli.get_connection_port(
            sp=self._sp, port_id=self._port_id, vport_id=self._vport_id,
            poll=self.poll)

    def property_names(self):
        names = super(VNXConnectionPort, self).property_names()
        names.append('type')
        return names

    def _get_virtual_port_prop(self, prop_name):
        if len(self.virtual_ports) == 0:
            return None
        elif len(self.virtual_ports) == 1:
            return getattr(self.virtual_ports[0], prop_name)
        else:
            return getattr(self.virtual_ports, prop_name)

    @property
    def virtual_port_id(self):
        return self._get_virtual_port_prop('virtual_port_id')

    @property
    def vport_id(self):
        return self.virtual_port_id

    @property
    def vlan_id(self):
        return self._get_virtual_port_prop('vlan_id')

    @property
    def ip_address(self):
        return self._get_virtual_port_prop('ip_address')

    @property
    def subnet_mask(self):
        return self._get_virtual_port_prop('subnet_mask')

    @property
    def gateway_address(self):
        return self._get_virtual_port_prop('gateway_address')

    @property
    def initiator_authentication(self):
        return self._get_virtual_port_prop('initiator_authentication')

    @classmethod
    def get(cls, cli, sp=None, port_id=None, vport_id=None, port_type=None,
            has_ip=None):
        VNXPortType.verify(port_type)
        if sp is not None and port_id is not None and vport_id is not None:
            ret = VNXConnectionPort(sp, port_id, vport_id, cli)
            if port_type is not None and ret.type != port_type:
                ret = None

        else:
            ret = VNXConnectionPortList(cli, sp, port_id, vport_id, port_type,
                                        has_ip)
        return ret


class VNXConnectionVirtualPort(VNXPort):
    def __init__(self, sp=None, port_id=None, vport_id=None, cli=None):
        super(VNXConnectionVirtualPort, self).__init__()
        self._sp = sp
        self._port_id = port_id
        self._vport_id = vport_id
        self._cli = cli

    @property
    def vport_id(self):
        return self.virtual_port_id

    def property_names(self):
        ret = super(VNXConnectionVirtualPort, self).property_names()
        ret.extend(['auto_negotiate', 'available_speed', 'current_mtu',
                    'enode_mac_address', 'flow_control', 'host_window',
                    'iscsi_alias', 'port_id', 'port_speed',
                    'replication_window', 'sp', 'wwn', 'type'])
        return ret

    def __hash__(self):
        return hash('<VNXPort {{sp: {}, port_id: {}, vport_id: {}}}'
                    .format(self.sp, self.port_id, self.vport_id))


class VNXConnectionVirtualPortList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXConnectionVirtualPort


class VNXStorageGroupHBA(VNXPort):
    @property
    def sp(self):
        return VNXSPEnum.parse(self.hba[1])

    @property
    def port_id(self):
        return int(self.hba[2])

    @property
    def uid(self):
        return self.hba[0]

    # @property
    # def vlan(self):
    #     sp_port = self.sp_port
    #     ret = None
    #
    #     if self.type == VNXPortType.ISCSI:
    #         # vport only valid for iSCSI
    #         if sp_port is not None and 'v' in sp_port:
    #             ret = int(sp_port[sp_port.find('v') + 1:])
    #     return ret

    @property
    def vport_id(self):
        sp_port = self.sp_port
        ret = None

        if self.type == VNXPortType.ISCSI:
            # vport only valid for iSCSI
            if sp_port is not None and 'v' in sp_port:
                ret = int(sp_port[sp_port.find('v') + 1:])
        return ret

    @property
    def port_type(self):
        return self.type

    @property
    def type(self):
        ret = None
        if '.' in self.uid:
            ret = VNXPortType.ISCSI
        elif ':' in self.uid:
            ret = VNXPortType.FC
        return ret

    def property_names(self):
        ret = super(VNXStorageGroupHBA, self).property_names()
        return ret + ['uid', 'sp', 'port_id', 'vport_id', 'vlan', 'port_type']


class VNXStorageGroupHBAList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXStorageGroupHBA



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

from hamcrest import assert_that, equal_to, raises

from storops_test.vnx.nas_mock import t_nas, patch_post

from storops.exception import VNXFsSnapNameInUseError
from storops.vnx.resource.fs_snap import VNXFsSnap

__author__ = 'Jay Xu'


class VXNFsSnapTest(unittest.TestCase):
    @staticmethod
    def verify_snap_230(snap):
        assert_that(snap.existed, equal_to(True))
        assert_that(snap.name, equal_to('ESA'))
        assert_that(snap.state, equal_to('active'))
        assert_that(snap.fs_id, equal_to(222))
        assert_that(snap.snap_id, equal_to(230))
        assert_that(snap.mover_id, equal_to(1))
        assert_that(snap.is_vdm, equal_to(False))

    @patch_post
    def test_get_all(self):
        snap_list = VNXFsSnap.get(t_nas())
        snap = next(snap for snap in snap_list if snap.snap_id == 230)
        self.verify_snap_230(snap)

    @patch_post
    def test_get_by_name(self):
        snap = VNXFsSnap.get(name='ESA', cli=t_nas())
        self.verify_snap_230(snap)

    @patch_post
    def test_get_by_name_not_found(self):
        snap = VNXFsSnap(name='aaa', cli=t_nas())
        assert_that(snap.existed, equal_to(False))

    @patch_post
    def test_get_by_id_not_found(self):
        snap = VNXFsSnap.get(snap_id=111, cli=t_nas())
        assert_that(snap.existed, equal_to(False))

    @patch_post
    def test_get_by_id(self):
        snap = VNXFsSnap(snap_id=230, cli=t_nas())
        self.verify_snap_230(snap)

    @patch_post
    def test_create_success(self):
        snap = VNXFsSnap.create(t_nas(), 'test', 222, 61)
        assert_that(snap.name, equal_to('test'))
        assert_that(snap.fs_id, equal_to(222))
        assert_that(snap.snap_id, equal_to(242))

    @patch_post
    def test_create_existed(self):
        def f():
            VNXFsSnap.create(t_nas(), 'Tan_Manual_CheckPoint', 228, 61)

        assert_that(f, raises(VNXFsSnapNameInUseError, 'already in use'))

    @patch_post
    def test_delete(self):
        snap = VNXFsSnap(name='test', cli=t_nas())
        resp = snap.delete()
        assert_that(resp.is_ok(), equal_to(True))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

import storops.unity.resource.cifs_server
import storops.unity.resource.dns_server
import storops.unity.resource.interface
import storops.unity.resource.nfs_server
import storops.unity.resource.pool
from storops.exception import UnityCifsServiceNotEnabledError, \
    UnityPolicyInvalidParametersError
from storops.unity.enums import ReplicationEndpointResourceTypeEnum
from storops.unity.resource import UnityResource, UnityResourceList
from storops.unity.resource.replication_session import UnityResourceConfig, \
    UnityReplicationSession
from storops.unity.resource.sp import UnityStorageProcessor
from storops.unity.resource.tenant import UnityTenant

__author__ = 'Jay Xu'

log = logging.getLogger(__name__)


class UnityNasServer(UnityResource):
    @classmethod
    def create(cls, cli, name, sp, pool, is_repl_dst=None,
               multi_proto=None,
               tenant=None):
        sp = UnityStorageProcessor.get(cli, sp)
        pool_clz = storops.unity.resource.pool.UnityPool
        pool = pool_clz.get(cli, pool)
        if tenant is not None:
            tenant = UnityTenant.get(cli, tenant)

        resp = cli.post(cls().resource_class,
                        name=name,
                        homeSP=sp,
                        pool=pool,
                        isReplicationDestination=is_repl_dst,
                        isMultiProtocolEnabled=multi_proto,
                        tenant=tenant)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    def delete(self, skip_domain_unjoin=None, username=None,
               password=None, async_mode=False):
        resp = self._cli.delete(self.resource_class,
                                self.get_id(),
                                skipDomainUnjoin=skip_domain_unjoin,
                                domainUsername=username,
                                domainPassword=password,
                                async_mode=async_mode)
        resp.raise_if_err()
        return resp

    def create_file_interface(self, ip_port, ip, netmask=None, gateway=None,
                              vlan_id=None, role=None,
                              v6_prefix_length=None):
        clz = storops.unity.resource.interface.UnityFileInterface
        return clz.create(self._cli, self,
                          ip_port=ip_port, ip=ip, netmask=netmask,
                          gateway=gateway, vlan_id=vlan_id, role=role,
                          v6_prefix_length=v6_prefix_length)

    def create_cifs_server(self, interfaces=None,
                           netbios_name=None, name=None,
                           domain=None, domain_username=None,
                           domain_password=None,
                           workgroup=None, local_password=None):
        clz = storops.unity.resource.cifs_server.UnityCifsServer
        return clz.create(self._cli, self,
                          interfaces=interfaces,
                          netbios_name=netbios_name,
                          name=name,
                          domain=domain,
                          domain_username=domain_username,
                          domain_password=domain_password,
                          workgroup=workgroup,
                          local_password=local_password)

    def enable_cifs_service(self, interfaces=None,
                            netbios_name=None, name=None,
                            domain=None, domain_username=None,
                            domain_password=None,
                            workgroup=None, local_password=None
                            ):
        if domain_username is not None and domain is None:
            dns_server = self.file_dns_server
            if dns_server is not None:
                domain = dns_server.domain
        self.create_cifs_server(interfaces=interfaces,
                                netbios_name=netbios_name,
                                name=name,
                                domain=domain,
                                domain_username=domain_username,
                                domain_password=domain_password,
                                workgroup=workgroup,
                                local_password=local_password)

    def create_nfs_server(self, host_name=None, nfs_v4_enabled=True,
                          kdc_type=None, kdc_username=None, kdc_password=None):
        clz = storops.unity.resource.nfs_server.UnityNfsServer
        return clz.create(self._cli, self,
                          host_name=host_name,
                          nfs_v4_enabled=nfs_v4_enabled,
                          kdc_type=kdc_type,
                          kdc_username=kdc_username,
                          kdc_password=kdc_password)

    def enable_nfs_service(self, host_name=None, nfs_v4_enabled=True,
                           kdc_type=None, kdc_username=None,
                           kdc_password=None):
        self.create_nfs_server(host_name=host_name,
                               nfs_v4_enabled=nfs_v4_enabled,
                               kdc_type=kdc_type,
                               kdc_username=kdc_username,
                               kdc_password=kdc_password)

    def create_dns_server(self, domain, *ip_list):
        clz = storops.unity.resource.dns_server.UnityFileDnsServer
        return clz.create(self._cli, self, domain=domain, ip_list=ip_list)

    def get_cifs_server(self):
        cifs_server_list = self.cifs_server
        if cifs_server_list:
            ret = cifs_server_list[0]
        else:
            raise UnityCifsServiceNotEnabledError(
                'CIFS is not enabled on {}.'.format(self.name))
        return ret

    def replicate(self, dst_nas_server_id, max_time_out_of_sync,
                  remote_system=None, replication_name=None):
        """
        Creates a replication session with an existing nas server as
        destination.

        :param dst_nas_server_id: destination nas server id.
        :param max_time_out_of_sync: maximum time to wait before syncing the
            source and destination. Value `-1` means the automatic sync is not
            performed. `0` means it is a sync replication.
        :param remote_system: `UnityRemoteSystem` object. The remote system to
            which the replication is being configured. When not specified, it
            defaults to local system.
        :param replication_name: replication name.
        :return: created replication session.
        """

        return UnityReplicationSession.create(
            self._cli, self.get_id(), dst_nas_server_id, max_time_out_of_sync,
            name=replication_name, remote_system=remote_system)

    def replicate_with_dst_resource_provisioning(self, max_time_out_of_sync,
                                                 dst_pool_id,
                                                 dst_nas_server_name=None,
                                                 remote_system=None,
                                                 replication_name=None,
                                                 dst_sp=None,
                                                 is_backup_only=None):
        """
        Creates a replication session with destination nas server provisioning.

        :param max_time_out_of_sync: maximum time to wait before syncing the
            source and destination. Value `-1` means the automatic sync is not
            performed. `0` means it is a sync replication.
        :param dst_pool_id: id of pool to allocate destination nas server.
        :param dst_nas_server_name: destination nas server name.
        :param remote_system: `UnityRemoteSystem` object. The remote system to
            which the replication is being configured. When not specified, it
            defaults to local system.
        :param replication_name: replication name.
        :param dst_sp: `NodeEnum` value. Default storage processor of
            destination nas server. It is required to create remote
            replication.
        :param is_backup_only: is backup only or not.
        :return: created replication session.
        """
        if remote_system and (dst_sp is None):
            message = 'Default storage processor is required to create ' \
                      'replication session with remote Unity system.'
            raise UnityPolicyInvalidParametersError(message)
        dst_resource = UnityResourceConfig.to_embedded(
            name=dst_nas_server_name, pool_id=dst_pool_id,
            default_sp=dst_sp, is_backup_only=is_backup_only,
            replication_resource_type=(
                ReplicationEndpointResourceTypeEnum.NASSERVER))
        return UnityReplicationSession.create_with_dst_resource_provisioning(
            self._cli, self.get_id(), dst_resource, max_time_out_of_sync,
            remote_system=remote_system, name=replication_name,
        )

    def modify(self, name=None, sp=None, is_replication_destination=None,
               is_backup_only=None, current_unix_directory_service=None,
               is_multi_protocol_enabled=None, allow_unmapped_user=None,
               default_unix_user=None, default_windows_user=None,
               enable_windows_to_unix_username=None,
               is_packet_reflect_enabled=None, is_ignore_warnings=None):
        """
        Modify a NAS server.

        :param name: specify a new NAS server name.
        :param sp: specify the SP on which the VDM is to run.
        :param is_replication_destination: specify whether the NAS server
            is a replication destination. Values are:
            true - Replication destination NAS server.
            false - Normal NAS server.
        :param is_backup_only: specify whether the NAS server is used as
            backup only. Values are:
            true - NAS server acts as backup only.
            false - Normal NAS server.
        :param current_unix_directory_service: directory Service used for
            quering identity information for UNIX (such as UIDs, GIDs, net
            groups).
        :param is_multi_protocol_enabled: indicates whether multiprotocol
            sharing mode is enabled. Values are:
            true - Enable multiprotocol sharing.
            false - Disable multiprotocol sharing.
        :param allow_unmapped_user: use this flag to mandatory disable access
            in case of any user mapping failure. Values are:
            true - Enable access in case of any user mapping failure.
            false - Disable access in case of any user mapping failure.
        :param default_unix_user: default Unix user name used for granting
            access in case of Windows to Unix user mapping failure.
            When empty, access in such case is denied.
        :param default_windows_user: default Windows user name used for
            granting access in case of Unix to Windows user mapping failure.
            When empty, access in such case is denied.
        :param enable_windows_to_unix_username: indicates whether a
            Unix to/from Windows user name mapping is enabled. Values are:
            true - Unix to/from Windows user name mapping is enabled.
            false - Unix to/from Windows user name mapping is disabled.
        :param is_packet_reflect_enabled.
        :param is_ignore_warnings: is backup only or not.
        """
        req_body = self._cli.make_body(
            name=name, homeSP=sp,
            isReplicationDestination=is_replication_destination,
            isBackupOnly=is_backup_only,
            currentUnixDirectoryService=current_unix_directory_service,
            isMultiProtocolEnabled=is_multi_protocol_enabled,
            allowUnmappedUser=allow_unmapped_user,
            defaultUnixUser=default_unix_user,
            defaultWindowsUser=default_windows_user,
            enableWindowsToUnixUsernameMapping=enable_windows_to_unix_username,
            isPacketReflectEnabled=is_packet_reflect_enabled,
            isIgnoreWarnings=is_ignore_warnings)

        resp = self.action('modify', **req_body)
        resp.raise_if_err()
        return resp


class UnityNasServerList(UnityResourceList):
    def __init__(self, cli=None, home_sp=None, current_sp=None, **filters):
        super(UnityNasServerList, self).__init__(cli, **filters)
        self._home_sp_id = None
        self._current_sp_id = None
        self._set_filter(home_sp, current_sp)

    def _set_filter(self, home_sp=None, current_sp=None, **kwargs):
        self._home_sp_id, self._current_sp_id = (
            [sp.get_id() if isinstance(sp, UnityStorageProcessor)
             else sp for sp in (home_sp, current_sp)])

    def _filter(self, nas_server):
        ret = True
        if self._home_sp_id is not None:
            ret &= nas_server.home_sp.get_id() == self._home_sp_id
        if self._current_sp_id is not None:
            ret &= nas_server.current_sp.get_id() == self._current_sp_id
        return ret

    @classmethod
    def get_resource_class(cls):
        return UnityNasServer



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

from storops.vnx.resource.fs import VNXFileSystem
from storops.vnx.resource.mover import VNXMover
from storops.vnx.resource import VNXResource, VNXCliResourceList

__author__ = 'Jay Xu'

log = logging.getLogger(__name__)


class VNXFsMountPointList(VNXCliResourceList):
    def __init__(self, mover=None, path=None, cli=None):
        super(VNXFsMountPointList, self).__init__(cli=cli)
        self._mover = mover
        self._path = path

    @classmethod
    def get_resource_class(cls):
        return VNXFsMountPoint

    def _get_raw_resource(self):
        if self._mover:
            mover_id = self._mover.get_mover_id()
            is_vdm = self._mover.is_vdm
        else:
            mover_id = None
            is_vdm = None
        return self._cli.get_fs_mp(
            mover_id=mover_id, is_vdm=is_vdm, path=self._path)


class VNXFsMountPoint(VNXResource):
    def __init__(self, mover=None, path=None, cli=None):
        super(VNXFsMountPoint, self).__init__()
        self._path = path
        self._mover = mover
        self._cli = cli

    def _get_raw_resource(self):
        if self._mover is None:
            raise ValueError('mover not specified.')
        elif self._path is None:
            raise ValueError('path not specified.')
        return self._cli.get_fs_mp(path=self._path,
                                   mover_id=self._mover.get_mover_id(),
                                   is_vdm=self._mover.is_vdm)

    def get_path(self):
        if self._path is not None:
            ret = self._path
        else:
            ret = self.path
        return ret

    @classmethod
    def create(cls, cli, path, fs, mover):
        fs_id = VNXFileSystem.get_id(fs)
        resp = cli.create_fs_mp(path, fs_id, mover.get_mover_id(),
                                mover.is_vdm)
        resp.raise_if_err()
        return VNXFsMountPoint(mover, path, cli)

    def delete(self):
        resp = self._cli.delete_fs_mp(self.get_path(),
                                      self.mover.get_mover_id(),
                                      self.mover.is_vdm)
        resp.raise_if_err()
        return resp

    @property
    def mover(self):
        if self._mover is not None:
            ret = self._mover
        else:
            ret = VNXMover(mover_id=self.mover_id, cli=self._cli)
        return ret

    @property
    def filesystem(self):
        return VNXFileSystem(fs_id=self.fs_id, cli=self._cli)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import re
import six

from storops.lib.common import cache
from storops.lib.parser import ParserConfigFactory, OutputParser

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class UnityParserConfigFactory(ParserConfigFactory):
    @classmethod
    def get_parser_clz(cls, data_src):
        if data_src is None or data_src == 'rest':
            ret = UnityRestParser
        else:
            raise ValueError('data_src {} not supported.'.format(data_src))
        return ret


_factory_singleton = UnityParserConfigFactory()


@cache
def get_unity_parser(name):
    return _factory_singleton.get(name)


class UnityRestParser(OutputParser):
    data_src = 'rest'

    def __init__(self):
        super(UnityRestParser, self).__init__()
        self.name = None

    def parse_all(self, output, properties=None):
        try:
            output = output.contents
        except AttributeError:
            pass
        return output

    def parse(self, output, properties=None):
        try:
            output = output.first_content
        except AttributeError:
            pass
        return self._parse_object(output, properties=None,
                                  preloaded_props=properties)

    def _parse_object(self, obj, properties=None, preloaded_props=None):
        if properties is None:
            properties = self.properties

        ret = {}
        for p in properties:
            if isinstance(obj, list):
                log.error('cannot parse list: {}.  '
                          'a list converter must be specified.'.format(obj))
                continue
            if p.label in obj.keys():
                value = p.convert(obj[p.label])
                if preloaded_props is not None and isinstance(
                        preloaded_props, NestedProperties):
                    subtree = preloaded_props.get_child_subtree(p.key)
                    if (subtree is not None and
                            hasattr(value, 'set_preloaded_properties')):
                        value.set_preloaded_properties(subtree)
                ret[p.key] = value
        return ret

    def init_from_config(self, config):
        self.name = config.name


class NestedProperty(object):
    def __init__(self, key):
        self.key = key

    @property
    def label(self):
        return self.under_score_to_camel_case(self.key)

    @classmethod
    def under_score_to_camel_case(cls, value):
        ret = re.sub(r'_([a-z])', lambda a: a.group(1).upper(), value)
        return ret

    def get_first_level_key(self):
        return self.key.split('.')[0]

    def remove_first_level_key(self):
        pos = self.key.find('.')
        if pos >= 0:
            return self.key[pos + 1:]
        else:
            return None


class NestedProperties(object):
    def __init__(self, *keys):
        self._props = map(NestedProperty, keys)
        self._map = None
        self._query_fields = None

    @classmethod
    def build(cls, properties):
        ret = None
        if not properties:
            ret = None
        elif isinstance(properties, six.text_type):
            ret = NestedProperties(properties)
        elif isinstance(properties, (list, tuple, set)):
            ret = NestedProperties(*properties)
        else:
            log.error('invalid properties {} to build NestedProperties '
                      'object.'.format(properties))
        return ret

    @property
    def _prop_map(self):
        if not self._map:
            map = {}
            for p in self._props:
                key = p.get_first_level_key()
                child_prop = p.remove_first_level_key()
                if child_prop is not None:
                    map.setdefault(key, []).append(child_prop)
                else:
                    map.setdefault(key, [])
            self._map = map
        return self._map

    def get_properties(self):
        return self._prop_map.keys()

    def get_child_subtree(self, prop):
        if prop not in self._prop_map:
            return None
        if len(self._prop_map[prop]) == 0:
            return None
        return NestedProperties.build(self._prop_map[prop])

    @property
    def query_fields(self):
        if self._query_fields is None:
            self._query_fields = [a.label for a in self._props]
        return tuple(self._query_fields)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to

from storops.vnx.nas_cmd import NasCommand
from storops.vnx.resource.cifs_share import CifsAccessControl

__author__ = 'Cedric Zhuang'


class NasCommandTest(TestCase):
    def setUp(self):
        self.cmd = NasCommand()

    def test_nas_cel_list(self):
        cmd = ' '.join(self.cmd.nas_cel_list())
        assert_that(cmd, equal_to('/nas/bin/nas_cel -interconnect -l'))

    def test_get_dm_interfaces_all_mover(self):
        cmd = ' '.join(self.cmd.get_dm_interfaces(is_vdm=False))
        assert_that(cmd, equal_to('/nas/bin/nas_server -i -all'))

    def test_get_dm_interfaces_all_vdm(self):
        cmd = ' '.join(self.cmd.get_dm_interfaces(is_vdm=True))
        assert_that(cmd, equal_to('/nas/bin/nas_server -i -vdm -all'))

    def test_get_dm_interfaces_mover_name(self):
        cmd = ' '.join(
            self.cmd.get_dm_interfaces(name='server_2', is_vdm=False))
        assert_that(cmd, equal_to('/nas/bin/nas_server -i server_2'))

    def test_get_dm_interfaces_vdm_name(self):
        cmd = ' '.join(self.cmd.get_dm_interfaces(name='VDM_ESA', is_vdm=True))
        assert_that(cmd, equal_to('/nas/bin/nas_server -i -vdm VDM_ESA'))

    def test_attach_nfs_interface(self):
        cmd = ' '.join(
            self.cmd.attach_nfs_interface('1.1.1.1-0', vdm_name='my'))
        assert_that(cmd,
                    equal_to('/nas/bin/nas_server -vdm my -attach 1.1.1.1-0'))

    def test_detach_nfs_interface(self):
        cmd = ' '.join(
            self.cmd.detach_nfs_interface('1.1.1.1-0', vdm_name='my'))
        assert_that(cmd,
                    equal_to('/nas/bin/nas_server -vdm my -detach 1.1.1.1-0'))

    def test_disable_share_access(self):
        cmd = ' '.join(self.cmd.disable_cifs_share_access('zzz', 'server_2'))
        expected = ('/nas/bin/.server_config server_2 -v sharesd zzz '
                    'set noaccess')
        assert_that(cmd, equal_to(expected))

    def test_allow_share_access(self):
        cmd = ' '.join(self.cmd.allow_cifs_share_access(
            'zzz', 'server_2', 'admin', 'a.dev', CifsAccessControl.READ))
        expected = ('/nas/bin/.server_config server_2 -v sharesd zzz '
                    'grant admin@a.dev=read')
        assert_that(cmd, equal_to(expected))

    def test_deny_share_access(self):
        cmd = ' '.join(self.cmd.deny_cifs_share_access(
            'zzz', 'server_2', 'admin', 'a.dev'))
        expected = ('/nas/bin/.server_config server_2 -v sharesd zzz '
                    'revoke admin@a.dev=fullcontrol')
        assert_that(cmd, equal_to(expected))



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

import retryz

import storops.unity.resource.move_session
import storops.unity.resource.pool
from storops.exception import UnityBaseHasThinCloneError, \
    UnityResourceNotFoundError, UnityCGMemberActionNotSupportError, \
    UnityThinCloneNotAllowedError, UnityMigrationSourceHasThinCloneError, \
    UnityMigrationTimeoutException
from storops.lib import job_helper
from storops.lib.thinclone_helper import TCHelper
from storops.lib.version import version
from storops.unity import enums
from storops.unity.client import UnityClient
from storops.unity.enums import TieringPolicyEnum, NodeEnum, \
    HostLUNAccessEnum, ThinCloneActionEnum, StorageResourceTypeEnum
from storops.unity.resource import UnityResource, UnityResourceList
from storops.unity.resource.host import UnityHostList
from storops.unity.resource.replication_session import UnityResourceConfig, \
    UnityReplicationSession
from storops.unity.resource.snap import UnitySnap, UnitySnapList
from storops.unity.resource.sp import UnityStorageProcessor
from storops.unity.resource.storage_resource import UnityStorageResource
from storops.unity.resp import RESP_OK

__author__ = 'Jay Xu'

log = logging.getLogger(__name__)


def prepare_lun_parameters(cli=None, **kwargs):
    @version('<4.3')
    def make_compression_body(cli=None,
                              is_compression=None):
        return UnityClient.make_body(isCompressionEnabled=is_compression)

    @version('>=4.3')  # noqa
    def make_compression_body(cli=None,
                              is_compression=None):
        return UnityClient.make_body(isDataReductionEnabled=is_compression)

    sp = kwargs.get('sp')
    if isinstance(sp, UnityStorageProcessor):
        sp_node = sp.to_node_enum()
    elif isinstance(sp, NodeEnum):
        sp_node = sp
    else:
        sp_node = NodeEnum.parse(sp)
    NodeEnum.verify(sp_node)

    TieringPolicyEnum.verify(kwargs.get('tiering_policy'))

    lun_parameters = UnityClient.make_body(
        isThinEnabled=kwargs.get('is_thin'),
        size=kwargs.get('size'),
        pool=kwargs.get('pool'),
        defaultNode=sp_node,
        fastVPParameters=UnityClient.make_body(
            tieringPolicy=kwargs.get('tiering_policy')),
        ioLimitParameters=UnityClient.make_body(
            ioLimitPolicy=kwargs.get('io_limit_policy')))

    compression_body = make_compression_body(
        cli,
        kwargs.get('is_compression'))

    lun_parameters.update(compression_body)

    # Empty host access can be used to wipe the host_access
    host_access = UnityClient.make_body(kwargs.get('host_access'),
                                        allow_empty=True)

    if host_access is not None:
        lun_parameters['hostAccess'] = host_access
    return lun_parameters


class UnityLun(UnityResource):
    _is_cg_member = None
    _cg = None
    _is_vmware_vmfs = None

    @classmethod
    def get_nested_properties(cls):
        return (
            'pool.name',
            'pool.raid_type',
            'pool.isFASTCacheEnabled',
            'host_access.host.name',
            'storage_resource.type'  # To avoid query parent type
        )

    @classmethod
    def create(cls, cli, name, pool, size, sp=None, host_access=None,
               is_thin=None, description=None, io_limit_policy=None,
               is_repl_dst=None, tiering_policy=None, snap_schedule=None,
               is_snap_schedule_paused=None, skip_sync_to_remote_system=None,
               is_compression=None, create_vmfs=False, major_version=None,
               block_size=None):
        pool_clz = storops.unity.resource.pool.UnityPool
        pool = pool_clz.get(cli, pool)

        req_body = cls._compose_lun_parameter(
            cli, name=name, pool=pool, size=size, sp=sp, is_thin=is_thin,
            host_access=host_access, description=description,
            io_limit_policy=io_limit_policy, is_repl_dst=is_repl_dst,
            tiering_policy=tiering_policy, snap_schedule=snap_schedule,
            is_snap_schedule_paused=is_snap_schedule_paused,
            skip_sync_to_remote_system=skip_sync_to_remote_system,
            is_compression=is_compression, major_version=major_version,
            block_size=block_size)

        create_method = 'createVmwareLun' if create_vmfs else 'createLun'

        resp = cli.type_action(UnityStorageResource().resource_class,
                               create_method, **req_body)
        resp.raise_if_err()
        sr = UnityStorageResource(_id=resp.resource_id, cli=cli)
        return sr.luns[0]

    @property
    def name(self):
        if hasattr(self, '_name') and self._name is not None:
            name = self._name
        else:
            if not self._is_updated():
                self.update()
            name = self._get_value_by_key('name')
        return name

    @name.setter
    def name(self, new_name):
        self.modify(name=new_name)

    @property
    def is_cg_member(self):
        if self._is_cg_member is None:  # None means unknown, requires a query
            return (self.storage_resource.type ==
                    StorageResourceTypeEnum.CONSISTENCY_GROUP)
        else:
            return self._is_cg_member

    @is_cg_member.setter
    def is_cg_member(self, is_cg_member):
        self._is_cg_member = is_cg_member

    @property
    def cg(self):
        if self.is_cg_member and self._cg is None:
            from storops.unity.resource.cg import UnityConsistencyGroup
            self._cg = UnityConsistencyGroup(cli=self._cli,
                                             _id=self.storage_resource.id)
        return self._cg

    @property
    def io_limit_rule(self):
        rule = None
        if self.io_limit_policy:
            policy = self.io_limit_policy
            if policy.io_limit_rule_settings:
                rules = policy.io_limit_rule_settings
            elif policy.io_limit_rules:
                rules = policy.io_limit_rules
            else:
                rules = None

            if rules:
                rule = rules[0]
        return rule

    @property
    def total_size_gb(self):
        return self.size_total / (1024 ** 3)

    @total_size_gb.setter
    def total_size_gb(self, value):
        self.expand(value * 1024 ** 3)

    @property
    def max_iops(self):
        return self.effective_io_limit_max_iops

    @property
    def max_kbps(self):
        return self.effective_io_limit_max_kbps

    @property
    def is_vmware_vmfs(self):
        # None means unknown, requires a query
        if self._is_vmware_vmfs is None:
            self._is_vmware_vmfs = (self.storage_resource.type ==
                                    StorageResourceTypeEnum.VMWARE_ISCSI)
        return self._is_vmware_vmfs

    def expand(self, new_size):
        """ expand the LUN to a new size

        :param new_size: new size in bytes.
        :return: the old size
        """
        ret = self.size_total
        resp = self.modify(size=new_size)
        resp.raise_if_err()
        return ret

    @staticmethod
    def _compose_lun_parameter(cli, **kwargs):

        body = cli.make_body(
            name=kwargs.get('name'),
            description=kwargs.get('description'),
            replicationParameters=cli.make_body(
                isReplicationDestination=kwargs.get('is_repl_dst')),
            vmwareIscsiParameters=cli.make_body(
                majorVersion=kwargs.get('major_version'),
                blockSize=kwargs.get('block_size')),
            snapScheduleParameters=cli.make_body(
                snapSchedule=kwargs.get('snap_schedule'),
                isSnapSchedulePaused=kwargs.get('is_snap_schedule_paused'),
                skipSyncToRemoteSystem=kwargs.get('skip_sync_to_remote_system')
            )
        )

        # `hostAccess` could be empty list which is used to remove all host
        # access
        lun_parameters = prepare_lun_parameters(cli, **kwargs)
        if lun_parameters:
            body['lunParameters'] = lun_parameters
        return body

    def modify(self, name=None, size=None, host_access=None,
               description=None, sp=None, io_limit_policy=None,
               is_repl_dst=None, tiering_policy=None, snap_schedule=None,
               is_snap_schedule_paused=None, skip_sync_to_remote_system=None,
               is_compression=None, major_version=None, block_size=None):
        if self.is_cg_member:
            if any(each is not None for each in [is_repl_dst, snap_schedule,
                                                 is_snap_schedule_paused,
                                                 skip_sync_to_remote_system]):
                log.warning('LUN in CG not support to modify `is_repl_dst`'
                            ' `snap_schedule`, `is_snap_schedule_paused` and'
                            ' `skip_sync_to_remote_system`.')
            return self.cg.modify_lun(self, name=name, size=size,
                                      host_access=host_access,
                                      description=description, sp=sp,
                                      io_limit_policy=io_limit_policy,
                                      tiering_policy=tiering_policy,
                                      is_compression=is_compression)

        else:
            req_body = self._compose_lun_parameter(
                self._cli, name=name, pool=None, size=size, sp=sp,
                host_access=host_access, description=description,
                io_limit_policy=io_limit_policy, is_repl_dst=is_repl_dst,
                tiering_policy=tiering_policy, snap_schedule=snap_schedule,
                is_snap_schedule_paused=is_snap_schedule_paused,
                skip_sync_to_remote_system=skip_sync_to_remote_system,
                is_compression=is_compression, major_version=major_version,
                block_size=block_size)

            if self.is_vmware_vmfs:
                resp = self._cli.action(UnityStorageResource().resource_class,
                                        self.storage_resource.get_id(),
                                        'modifyVmwareLun', **req_body)
            else:
                resp = self._cli.action(UnityStorageResource().resource_class,
                                        self.get_id(), 'modifyLun', **req_body)
            resp.raise_if_err()
            return resp

    def delete(self, async_mode=True, force_snap_delete=False,
               force_vvol_delete=False, async_timeout=600,
               async_interval=1):
        sr = self.storage_resource
        if not self.existed or sr is None:
            raise UnityResourceNotFoundError(
                'cannot find lun {}.'.format(self.get_id()))
        resp = self._cli.delete(sr.resource_class, sr.get_id(),
                                forceSnapDeletion=force_snap_delete,
                                forceVvolDeletion=force_vvol_delete,
                                async_mode=async_mode)

        try:
            resp.raise_if_err()
            if async_mode and resp.job.existed:
                jh = job_helper.get_job_helper(self._cli)
                resp.job = jh.wait_job(resp.job, async_timeout,
                                       async_interval)

        except UnityBaseHasThinCloneError:
            log.warning('cannot delete the lun: %s, because it is a base lun '
                        'of a thin-clone.', self.get_id())
            TCHelper.notify(self, ThinCloneActionEnum.BASE_LUN_DELETE)
            return RESP_OK

        if self.is_thin_clone:
            TCHelper.notify(self, ThinCloneActionEnum.TC_DELETE)

        return resp

    def _attach_to(self, host, access_mask, hlu):
        host_access = [{'host': host, 'accessMask': access_mask}]
        if hlu is not None:
            host_access[0]['hlu'] = hlu
        # If this lun has been attached to other host, don't overwrite it.
        if self.host_access:
            host_access += [{'host': item.host,
                             'accessMask': item.access_mask} for item
                            in self.host_access if host.id != item.host.id]

        resp = self.modify(host_access=host_access)
        resp.raise_if_err()
        log.debug('Notify TCHelper the attaching action of lun: %s.',
                  self.get_id())
        TCHelper.notify(self, ThinCloneActionEnum.LUN_ATTACH)
        return resp

    @version('<4.4.0')  # noqa
    def attach_to(self, host, access_mask=HostLUNAccessEnum.PRODUCTION):
        return self._attach_to(host, access_mask, None)

    @version('>=4.4.0')  # noqa
    def attach_to(self, host, access_mask=HostLUNAccessEnum.PRODUCTION,
                  hlu=None):
        return self._attach_to(host, access_mask, hlu)

    def detach_from(self, host):
        if self.host_access is None:
            return None

        if host is None:
            # Detach the lun from all hosts if `host` is None
            log.info('Detach lun - %s from all hosts.', self.get_id())
            new_access = []
        else:
            new_access = [{'host': item.host,
                           'accessMask': item.access_mask} for item
                          in self.host_access if host.id != item.host.id]
        resp = self.modify(host_access=new_access)
        resp.raise_if_err()
        return resp

    def update_hosts(self, host_names):

        """Primarily for puppet-unity use.

        Update the hosts for the lun if needed.

        :param host_names: specify the new hosts which access the LUN.
        """

        if self.host_access:
            curr_hosts = [access.host.name for access in self.host_access]
        else:
            curr_hosts = []

        if set(curr_hosts) == set(host_names):
            log.info('Hosts for updating is equal to current hosts, '
                     'skip modification.')
            return None

        new_hosts = [UnityHostList.get(cli=self._cli, name=host_name)[0]
                     for host_name in host_names]
        new_access = [{'host': item,
                       'accessMask': HostLUNAccessEnum.PRODUCTION}
                      for item in new_hosts]
        resp = self.modify(host_access=new_access)
        resp.raise_if_err()
        return resp

    def create_snap(self, name=None, description=None, is_auto_delete=None,
                    retention_duration=None):
        if self.is_cg_member:
            raise UnityCGMemberActionNotSupportError()
        return UnitySnap.create(self._cli, self.storage_resource,
                                name=name, description=description,
                                is_auto_delete=is_auto_delete,
                                retention_duration=retention_duration,
                                is_read_only=None, fs_access_type=None)

    @version(">=4.2")
    def thin_clone(self, name, io_limit_policy=None, description=None):
        if self.is_cg_member:
            raise UnityCGMemberActionNotSupportError()

        if not self.is_thin_enabled:
            raise UnityThinCloneNotAllowedError()

        return TCHelper.thin_clone(self._cli, self, name, io_limit_policy,
                                   description)

    def _is_move_session_supported(self, dest):
        if self.is_thin_clone:
            log.error('Not support move session, source lun is thin clone.')
            return False
        if self.is_data_reduction_enabled and not dest.is_all_flash:
            log.error('Not support move session, source lun is compressed, '
                      'but destination pool is not all flash pool.')
            return False
        return True

    def migrate(self, dest, **kwargs):
        if not self._is_move_session_supported(dest):
            return False

        interval = kwargs.pop('interval', 5)
        timeout = kwargs.pop('timeout', 1800)
        is_thin = kwargs.get('is_thin')
        is_compressed = kwargs.get('is_compressed')

        @retryz.retry(timeout=timeout, wait=interval,
                      on_return=lambda x: not isinstance(x, bool))
        def _do_check_move_session(move_session_id):
            move_session = clz.get(self._cli, _id=move_session_id)
            if move_session.state == enums.MoveSessionStateEnum.COMPLETED:
                return True
            if move_session.state in [enums.MoveSessionStateEnum.FAILED,
                                      enums.MoveSessionStateEnum.CANCELLED]:
                return False

        clz = storops.unity.resource.move_session.UnityMoveSession
        if is_compressed not in (True, False):
            is_compressed = self.is_data_reduction_enabled
        try:
            move_session = clz.create(self._cli, self, dest,
                                      is_data_reduction_applied=is_compressed,
                                      is_dest_thin=is_thin)
            return _do_check_move_session(move_session.id)
        except UnityMigrationSourceHasThinCloneError:
            log.error('Not support move session, source lun has thin clone.')
            return False
        except retryz.RetryTimeoutError:
            raise UnityMigrationTimeoutException()

    # `__getstate__` and `__setstate__` are used by Pickle.
    def __getstate__(self):
        return {'_id': self.get_id(), 'cli': self._cli}

    def __setstate__(self, state):
        self.__init__(**state)

    @property
    def snapshots(self):
        return UnitySnapList(cli=self._cli,
                             storage_resource=self.storage_resource)

    def replicate(self, dst_lun_id, max_time_out_of_sync,
                  replication_name=None, replicate_existing_snaps=None,
                  remote_system=None):
        """
        Creates a replication session with a existing lun as destination.

        :param dst_lun_id: destination lun id.
        :param max_time_out_of_sync: maximum time to wait before syncing the
            source and destination. Value `-1` means the automatic sync is not
            performed. `0` means it is a sync replication.
        :param replication_name: replication name.
        :param replicate_existing_snaps: whether to replicate existing snaps.
        :param remote_system: `UnityRemoteSystem` object. The remote system to
            which the replication is being configured. When not specified, it
            defaults to local system.
        :return: created replication session.
        """

        return UnityReplicationSession.create(
            self._cli, self.get_id(), dst_lun_id, max_time_out_of_sync,
            name=replication_name,
            replicate_existing_snaps=replicate_existing_snaps,
            remote_system=remote_system)

    def replicate_with_dst_resource_provisioning(self, max_time_out_of_sync,
                                                 dst_pool_id,
                                                 dst_lun_name=None,
                                                 remote_system=None,
                                                 replication_name=None,
                                                 dst_size=None, dst_sp=None,
                                                 is_dst_thin=None,
                                                 dst_tiering_policy=None,
                                                 is_dst_compression=None):
        """
        Creates a replication session with destination lun provisioning.

        :param max_time_out_of_sync: maximum time to wait before syncing the
            source and destination. Value `-1` means the automatic sync is not
            performed. `0` means it is a sync replication.
        :param dst_pool_id: id of pool to allocate destination lun.
        :param dst_lun_name: destination lun name.
        :param remote_system: `UnityRemoteSystem` object. The remote system to
            which the replication is being configured. When not specified, it
            defaults to local system.
        :param replication_name: replication name.
        :param dst_size: destination lun size.
        :param dst_sp: `NodeEnum` value. Default storage processor of
            destination lun.
        :param is_dst_thin: indicates whether destination lun is thin or not.
        :param dst_tiering_policy: `TieringPolicyEnum` value. Tiering policy of
            destination lun.
        :param is_dst_compression: indicates whether destination lun is
            compression enabled or not.
        :return: created replication session.
        """

        dst_size = self.size_total if dst_size is None else dst_size

        dst_resource = UnityResourceConfig.to_embedded(
            name=dst_lun_name, pool_id=dst_pool_id,
            size=dst_size, default_sp=dst_sp,
            tiering_policy=dst_tiering_policy, is_thin_enabled=is_dst_thin,
            is_compression_enabled=is_dst_compression)
        return UnityReplicationSession.create_with_dst_resource_provisioning(
            self._cli, self.get_id(), dst_resource, max_time_out_of_sync,
            remote_system=remote_system, name=replication_name)

    def remove_snap_schedule(self):
        return self.modify(is_snap_schedule_paused=False)


class UnityLunList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityLun



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from collections import OrderedDict
import logging
import re

import six

from storops.lib.common import Dict, cache
from storops.lib.parser import OutputParser, PropDescriptor, \
    ParserConfigFactory

log = logging.getLogger(__name__)


class VNXPropDescriptor(PropDescriptor):
    pass


class VNXParserConfigFactory(ParserConfigFactory):
    @classmethod
    def get_parser_clz(cls, data_src):
        if data_src == 'cli':
            ret = VNXCliParser
        elif data_src == 'xmlapi':
            ret = VNXXmlApiParser
        else:
            raise ValueError('data_src {} not supported.'.format(data_src))
        return ret

    @classmethod
    def get_converter(cls, value):
        from storops.vnx import converter
        if hasattr(converter, value):
            ret = getattr(converter, value)
        else:
            ret = None
        return ret


_factory_singleton = VNXParserConfigFactory()


@cache
def get_vnx_parser(name):
    return _factory_singleton.get(name)


class VNXCliParser(OutputParser):
    data_src = 'cli'

    def _split_by_index(self, output):
        instances = []

        index_descriptor = self.index_property
        if index_descriptor is not None:
            index_pattern = index_descriptor.index_pattern

            last_start = 0
            has_match = False
            if isinstance(output, bytes):
                output = output.decode("utf-8")
            for match in re.finditer(index_pattern, output):
                has_match = True
                start = match.start()
                if start == 0:
                    continue
                else:
                    instances.append(output[last_start:start])
                    last_start = start
            if has_match:
                instances.append(output[last_start:])
        else:
            instances = [output]
        return instances

    def parse_single(self, output, properties=None):
        if isinstance(output, six.string_types):
            output = output.strip()
            ret = Dict()

            if properties is None:
                properties = self.properties

            for p in properties:
                matched = re.search(p.pattern, output)

                matched_value = None
                if matched is not None:
                    if len(matched.groups()) == 1:
                        value = matched.group(1)
                        value = value.strip()
                    else:
                        value = matched.groups()
                    value = p.convert(value)
                    matched_value = value
                elif p.is_index:
                    # index must have a match, skip this invalid input
                    ret = Dict()
                    break
                ret[p.key] = matched_value
        else:
            ret = output
        return ret

    def parse_all(self, output, properties=None):
        if isinstance(output, six.string_types):
            output = output.strip()
            split_outputs = self._split_by_index(output)
            instances = self._parse_split_output(split_outputs, properties)
            instances = self._merge_instance_with_same_index(instances)
        else:
            instances = output
        return instances

    def _merge_instance_with_same_index(self, instances):
        def key_gen(instance):
            str_keys = []
            for index in indices:
                str_keys.append(
                    '{}: {}'.format(index.key, instance[index.key]))
            return ', '.join(str_keys)

        def update_map(instance):
            key = key_gen(instance)
            if key not in idx_inst_map:
                idx_inst_map[key] = instance
            else:
                existed = idx_inst_map[key]
                for k, v in six.iteritems(instance):
                    if k not in existed or v is not None:
                        existed[k] = v

        idx_inst_map = OrderedDict()
        indices = self.index_property_list
        list(map(update_map, instances))
        return list(idx_inst_map.values())

    def _parse_split_output(self, split_outputs, properties):
        ret = []
        for instance in split_outputs:
            parsed = self.parse_single(instance, properties)
            if len(parsed) > 0:
                ret.append(parsed)
        return ret

    def parse(self, output, properties=None):
        ret = self.parse_all(output, properties)

        if len(ret) == 0:
            ret = Dict()
        elif isinstance(ret, (tuple, list)):
            ret = ret[0]
        return ret


class VNXXmlApiParser(OutputParser):
    data_src = 'xmlapi'

    def parse_all(self, output, properties=None):
        if hasattr(output, 'objects'):
            output = output.objects
        return [self._parse_object(obj) for obj in output]

    def parse(self, output, properties=None):
        if hasattr(output, 'objects') and len(output.objects):
            ret = self._parse_object(output.first_object, properties)
        elif isinstance(output, (list, tuple)) and len(output) > 0:
            ret = self._parse_object(output[0], properties)
        elif isinstance(output, dict):
            ret = output
        else:
            ret = Dict()
        return ret

    def _parse_object(self, obj, properties=None):
        if properties is None:
            properties = self.properties

        ret = {}
        for p in properties:
            if p.label in obj.keys():
                value = p.convert(obj[p.label])
                ret[p.key] = value
        return ret



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import equal_to, assert_that, instance_of, raises

from storops.exception import UnityException, UnityIpAddressUsedError, \
    UnityResourceNotFoundError
from storops.unity.enums import IpProtocolVersionEnum, FileInterfaceRoleEnum
from storops.unity.resource.health import UnityHealth
from storops.unity.resource.interface import UnityFileInterface, \
    UnityFileInterfaceList, UnityPreferredInterfaceSettings, \
    UnityPreferredInterfaceSettingsList
from storops.unity.resource.nas_server import UnityNasServer
from storops.unity.resource.port import UnityIpPort
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'


class UnityFileInterfaceTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        fi = UnityFileInterface('if_16', cli=t_rest())
        assert_that(fi.existed, equal_to(True))
        assert_that(fi.nas_server, instance_of(UnityNasServer))
        assert_that(fi.ip_port, instance_of(UnityIpPort))
        assert_that(fi.health, instance_of(UnityHealth))
        assert_that(fi.ip_address, equal_to('10.244.220.120'))
        assert_that(fi.ip_protocol_version,
                    equal_to(IpProtocolVersionEnum.IPv4))
        assert_that(fi.netmask, equal_to('255.255.255.0'))
        assert_that(fi.gateway, equal_to('10.244.220.1'))
        assert_that(fi.mac_address, equal_to('00:60:16:5C:08:E1'))
        assert_that(fi.name, equal_to('8_FNM00151200215'))
        assert_that(fi.role, equal_to(FileInterfaceRoleEnum.PRODUCTION))
        assert_that(fi.is_preferred, equal_to(True))
        assert_that(fi.is_disabled, equal_to(False))

    @patch_rest
    def test_get_all(self):
        fi_list = UnityFileInterfaceList(cli=t_rest())
        assert_that(len(fi_list), equal_to(1))

    @patch_rest
    def test_create_error_nas_server_not_found(self):
        def f():
            UnityFileInterface.create(
                t_rest(), 'nas_1', 'spa_eth2', '1.1.1.1',
                role=FileInterfaceRoleEnum.PRODUCTION)

        assert_that(f, raises(UnityException, 'Cannot find'))

    @patch_rest
    def test_create_success(self):
        fi = UnityFileInterface.create(
            t_rest(), 'nas_2', 'spa_eth2', '1.1.1.1',
            role=FileInterfaceRoleEnum.PRODUCTION)
        assert_that(fi.ip_address, equal_to('1.1.1.1'))

    @patch_rest
    def test_ip_address_in_use(self):
        def f():
            UnityFileInterface.create(
                t_rest(), 'nas_2', 'spa_eth2', '1.1.1.2',
                role=FileInterfaceRoleEnum.PRODUCTION)

        assert_that(f, raises(UnityIpAddressUsedError, 'already in use'))

    @patch_rest
    def test_delete_success(self):
        fi = UnityFileInterface(_id='if_20', cli=t_rest())
        resp = fi.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_not_found(self):
        def f():
            fi = UnityFileInterface(_id='if_25', cli=t_rest())
            fi.delete()

        assert_that(f, raises(UnityResourceNotFoundError, 'does not exist'))


class UnityPreferredInterfaceSettingsTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        pis = UnityPreferredInterfaceSettings('preferred_if_2', cli=t_rest())
        assert_that(pis.existed, equal_to(True))
        assert_that(pis.nas_server, instance_of(UnityNasServer))

    @patch_rest
    def test_get_all(self):
        pis_list = UnityPreferredInterfaceSettingsList(cli=t_rest())
        assert_that(len(pis_list), equal_to(1))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.vnx.enums import VNXProvisionEnum

from storops.exception import VNXNotEnoughDiskAvailableError, \
    VNXLunNotFoundError, VNXDeleteLunError
from storops.vnx.resource.disk import VNXDiskList

from storops.lib.common import instance_cache
from storops.vnx.resource import VNXCliResource, VNXCliResourceList
import storops.vnx.resource.lun
from storops import exception as ex

__author__ = 'Cedric Zhuang'


class VNXPoolFeature(VNXCliResource):
    def __init__(self, cli=None):
        super(VNXPoolFeature, self).__init__()
        self._cli = cli

    def _get_raw_resource(self):
        return self._cli.get_pool_feature(poll=self.poll)


class VNXPoolList(VNXCliResourceList):
    def __init__(self, cli=None, system_lun_list=None):
        super(VNXPoolList, self).__init__(cli)
        self._system_lun_list = system_lun_list

    @classmethod
    def get_resource_class(cls):
        return VNXPool

    def _get_raw_resource(self):
        return self._cli.get_pool(poll=self.poll)

    def _get_resource_instance(self):
        ret = super(VNXPoolList, self)._get_resource_instance()
        ret.set_system_lun_list(self._system_lun_list)
        return ret


class VNXPool(VNXCliResource):
    def __init__(self, pool_id=None, name=None, cli=None,
                 system_lun_list=None):
        super(VNXPool, self).__init__()
        self._cli = cli
        self._pool_id = pool_id
        self._name = name
        self._system_lun_list = system_lun_list

    def get_pool_id(self):
        if self._pool_id is not None:
            ret = self._pool_id
        else:
            ret = self.pool_id
        return ret

    def _get_name_or_id(self):
        ret = {'name': None, 'pool_id': None}
        if self._pool_id is not None:
            ret['pool_id'] = self._pool_id
        else:
            ret['name'] = self._get_name()
        return ret

    def __setattr__(self, key, value):
        if key == 'name':
            self.rename(value)
            return
        super(VNXPool, self).__setattr__(key, value)

    def rename(self, new_name):
        if new_name is not None and new_name != self._name:
            ret = self._cli.modify_storage_pool(new_name=new_name,
                                                poll=self.poll,
                                                **self._get_name_or_id())
            ex.raise_if_err(ret, default=ex.VNXModifyPoolError)
            self._name = new_name

    @staticmethod
    def create(cli, name, disks, raid_type=None, ):
        if isinstance(disks, VNXDiskList):
            disks = sorted(disks.index)
        if not disks:
            raise VNXNotEnoughDiskAvailableError()
        ret = cli.create_pool(name, disks, raid_type)
        ex.raise_if_err(ret, default=ex.VNXCreatePoolError)
        return VNXPool(name=name, cli=cli)

    def clear(self):
        lun_clz = storops.vnx.resource.lun.VNXLun
        if self.luns:
            for lun in self.luns:
                lun_id = lun_clz.get_id(lun)
                lun = lun_clz(lun_id=lun_id, cli=self._cli)
                try:
                    lun.delete(force=True)
                except (VNXLunNotFoundError, VNXDeleteLunError):
                    # ignore delete error
                    pass

    def delete(self, force=False):
        if force:
            self.clear()
        ret = self._cli.delete_pool(poll=self.poll, **self._get_name_or_id())
        ex.raise_if_err(ret, default=ex.VNXDeletePoolError)

    @classmethod
    def get(cls, cli, pool_id=None, name=None, system_lun_list=None):
        if pool_id is None and name is None:
            ret = VNXPoolList(cli, system_lun_list)
        else:
            ret = VNXPool(pool_id, name, cli, system_lun_list)
        return ret

    def create_lun(self,
                   lun_name=None,
                   size_gb=1,
                   lun_id=None,
                   provision=None,
                   tier=None,
                   ignore_thresholds=None):
        pool = {}
        if self._pool_id is not None:
            pool['pool_id'] = self._pool_id
        else:
            pool['pool_name'] = self._get_name()
        ret = self._cli.create_pool_lun(
            lun_name=lun_name,
            lun_id=lun_id,
            size_gb=size_gb,
            provision=provision,
            tier=tier,
            ignore_thresholds=ignore_thresholds,
            poll=self.poll,
            **pool)
        ex.raise_if_err(ret, 'error creating lun.',
                        default=ex.VNXCreateLunError)
        lun_clz = storops.vnx.resource.lun.VNXLun
        ret = lun_clz(lun_id, lun_name, self._cli)

        if provision == VNXProvisionEnum.COMPRESSED:
            ret.enable_compression()
            ret.update()
        return ret

    @staticmethod
    def delete_lun(lun, delete_snapshots=False, force_detach=False):
        lun.delete(delete_snapshots, force_detach)

    def _get_raw_resource(self):
        return self._cli.get_pool(poll=self.poll, **self._get_name_or_id())

    @instance_cache
    def get_lun(self):
        if self._system_lun_list is not None:
            ret = self._system_lun_list.shadow_copy(pool=self)
        else:
            clz = storops.vnx.resource.lun.VNXLunList
            ret = clz(self._cli, pool=self)
        return ret

    @property
    def lun_list(self):
        return self.get_lun()

    def set_system_lun_list(self, system_lun_list):
        self._system_lun_list = system_lun_list


class VNXPoolTier(VNXCliResource):
    pass


class VNXPoolTierList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXPoolTier



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Yong Huang'

LOG = logging.getLogger(__name__)


class UnityMoveSession(UnityResource):
    @classmethod
    def create(cls, cli, source_storage_resource, destination_pool,
               source_member_lun=None, is_dest_thin=None,
               is_data_reduction_applied=None, priority=None):
        req_body = cls._compose_move_session_parameter(
            cli, source_storage_resource=source_storage_resource,
            destination_pool=destination_pool,
            source_member_lun=source_member_lun,
            is_dest_thin=is_dest_thin,
            is_data_reduction_applied=is_data_reduction_applied,
            priority=priority)
        resp = cli.post(cls().resource_class, **req_body)
        resp.raise_if_err()
        move_session = cls.get(cli, resp.resource_id)
        return move_session

    def modify(self, cli, priority=None):
        req_body = self._compose_move_session_parameter(cli,
                                                        priority=priority)
        resp = self.action('modify', **req_body)
        resp.raise_if_err()
        return resp

    def cancel(self):
        resp = self.action('cancel')
        resp.raise_if_err()
        return resp

    @staticmethod
    def _compose_move_session_parameter(cli, source_storage_resource=None,
                                        destination_pool=None,
                                        source_member_lun=None,
                                        is_dest_thin=None,
                                        is_data_reduction_applied=None,
                                        priority=None):
        req_body = cli.make_body(
            sourceStorageResource=source_storage_resource,
            destinationPool=destination_pool,
            sourceMemberLun=source_member_lun,
            isDestThin=is_dest_thin,
            isDataReductionApplied=is_data_reduction_applied,
            priority=priority)
        return req_body


class UnityMoveSessionList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityMoveSession
